//-------------------------------------------------------------------------
// a simple test program to receive interrupts from the 
// WINNT-PCICC32 driver from ARW Elektronik, Germany
//
// this example accesses dircetly the WINNT driver through 'ioctls'
//
// (c) 2000 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// started                                         AR           28.02.2001
// released                                        AR           17.03.2001
//

//-------------------------------------------------------------------------
// INCLUDES
//
#include <stdio.h>
#include <windows.h>
#include <winioctl.h>
#include "..\pcicc32\pciCC32.h"    // include the structure definitions and IOCTL codes

#define DEFDEVICENAME "\\\\.\\PCICC32:\\CC32_xx"  // a device name 'template'
#define MODULE_NUMBER 1                           // module number of CC32

#define MESSAGESTRING "A program to receive interrupts from the PCICC32 WINNT driver. (%s)\n"

HANDLE hHandle = INVALID_HANDLE_VALUE;  // I use a global handle to my device path

//-------------------------------------------------------------------------
// what the program provides
//
void hlpMsg(void)
{
	printf(MESSAGESTRING, __DATE__);
    printf("usage: pciCC32IrqReceiver [-?]\n");
	printf("       ? - get help\n");
	printf("\n");
	printf("This program addresses interface #%d\n", MODULE_NUMBER);
	printf("\n");
}

//-------------------------------------------------------------------------
// create a DeviceName out of DEFDEVICENAME and the device number
//
static char *DeviceName(int nIfcNum)
{
	static char buffer[255];
	char *ptr = buffer;

	strcpy(buffer, DEFDEVICENAME);

	while (*ptr) ptr++;
	
	do
	{
		ptr--;
	} while (*ptr != '_');  // step back until '-'

	ptr++;
	
	if (nIfcNum >= 10)      // add the interface number
	{
		*ptr++   = '1';
		nIfcNum -= 10;
	}

	*ptr++ = '0' + nIfcNum;
	*ptr = 0;

	// a little help for the users
    return buffer;
}

//-------------------------------------------------------------------------
// open my path to the interface
//
static HANDLE openPath(char *name)
{
    return (HANDLE)CreateFile( 
                     name, 
                     GENERIC_READ | GENERIC_WRITE, 
                     0, 
                     NULL, 
                     OPEN_EXISTING, 
                     FILE_ATTRIBUTE_NORMAL, 
                     NULL 
                     );
}

//-------------------------------------------------------------------------
// init the interface
//
int Init_InterfaceNT(int nIfcNum)
{
  int error = 0;

  // start the driver ------------------
  if (hHandle == INVALID_HANDLE_VALUE)
  {
		if ((hHandle = openPath(DeviceName(nIfcNum))) == (HANDLE)-1)
			return GetLastError();
  }

  return 0;
}


//-------------------------------------------------------------------------------------------
// enable interrupt for a pciada / CC32 combination
//
int EnableInterrupts(HANDLE hHandle)
{
	DWORD result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
	PCICC32_IRQ_CONTROL irqCntrl;

	irqCntrl.dwInterface = 0;  // don't bother at WINNT
	irqCntrl.wEnable     = 1;  // enable it

	result = DeviceIoControl(hHandle, PCICC32_CONTROL_INTERRUPTS,
					&irqCntrl, sizeof(irqCntrl), NULL, 0, &DIOC_count, NULL);	 

	if (!result)
	{
		result = GetLastError();
		printf("Can't enable interrupt! (0x%08x)\n", result);
		return 0;
	}	

	return 1;
}


//-------------------------------------------------------------------------------------------
// disable interrupt for a pciada / CC32 combination
//
int DisableInterrupts(HANDLE hHandle)
{
	DWORD result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
	PCICC32_IRQ_CONTROL irqCntrl;

	irqCntrl.dwInterface = 0;  // don't bother at WINNT
	irqCntrl.wEnable     = 0;  // disable it

	result = DeviceIoControl(hHandle, PCICC32_CONTROL_INTERRUPTS,
					&irqCntrl, sizeof(irqCntrl), NULL, 0, &DIOC_count, NULL);	 

	if (!result)
	{
		printf("Can't disable interrupt! (0x%08x)\n", result);
		return 0;
	}	

	return 1;
}

//-------------------------------------------------------------------------------------------
// read waiting a status from the interface
//
int BlockingReadStatus(HANDLE hHandle, unsigned long *dwStatus)
{
	DWORD result;
	DWORD	DIOC_count;     // count of returned bytes of DeviceIoControl
	PCICC32_IRQ_RESPONSE response;
	int nError;
	OVERLAPPED sOverlapped;

	sOverlapped.hEvent    = NULL;
	result = DeviceIoControl(hHandle, PCICC32_INSTALL_IRQ_BLOCK,
					NULL, 0, &response, sizeof(response), &DIOC_count, NULL);
	
	*dwStatus = response.dwInterruptFlags;

	if (!result) 
	{
		nError = GetLastError();

		// a real error
		if ((nError != ERROR_IO_PENDING) && (nError != ERROR_SUCCESS)) 
		{
			printf("Can't determine status! (0x%08x)\n", nError);
			return 0;
		}

		// my IO is pending - wait for it
		if (nError == ERROR_IO_PENDING)
		{
			result = GetOverlappedResult(hHandle, &sOverlapped, &DIOC_count, TRUE);
			
			*dwStatus = response.dwInterruptFlags;

			if (!result) 
			{
				nError = GetLastError();   // if it was cancelled
				if (nError != ERROR_OPERATION_ABORTED) 
					printf("Can't get status! (0x%08x)\n", nError);
				return 0;
			}
		}
	}

	return 1;
}


//-------------------------------------------------------------------------
// the main program
//
int main(int argc, char *argv[])
{
	int result;
	unsigned long dwStatus;

	if (argc != 1)
	{
		hlpMsg();
		exit(0);
	}

	printf(MESSAGESTRING, __DATE__);
	printf("Ctrl-C to stop.\n");

	result = Init_InterfaceNT(MODULE_NUMBER);  // open the device path to the 1st interface
	if (result)
	{
		printf("Can't Init_InterfaceNT() - returns error 0x%08x\n", result);
		return(result);
	}
	else
	{	
		// enable interrupts
		EnableInterrupts(hHandle);

		// get status
		result = 1;
		while (result)
		{
			result = BlockingReadStatus(hHandle, &dwStatus);

			printf("Got status 0x%08x.\n", dwStatus);

			if (result)
				EnableInterrupts(hHandle);
		}

		// disable interrupts
		DisableInterrupts(hHandle);

		// close device -------------------------------------
		if (hHandle != INVALID_HANDLE_VALUE)
			CloseHandle(hHandle);

		return 0;
	}
}