//-------------------------------------------------------------------------
// WINNT driver for PCICC32 interface from ARW Elektronik, Germany --------
// all around irq handling
//
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and 
// modify it unless this header   is not modified or deleted. No warranty is given that 
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what                                            who          when
// started                                         AR           01.08.1999
// first release 1.0							   AR			17.10.1999
// IoConnectInterrupt, share vector now true       AR           04.03.2000
// globalInterruptEnabledStatus() added            AR           24.02.2001
// added IRQ handling                              AR           24.02.2001
//
  
//-------------------------------------------------------------------------
// INCLUDES
//
#include <ntddk.h>
#include <pcicc32_drv.h>
#include <pcicc32_i.h>
#include <pcicc32.h>
#include <pcicc32_local.h>

//------------------------------------------------------------------------
// DEFINES
//
#ifndef DWORD
#define DWORD ULONG
#endif

#ifndef WORD
#define WORD USHORT
#endif

//------------------------------------------------------------------------
// GLOBALS
//

//------------------------------------------------------------------------
// FUNCTIONS
//

//------------------------------------------------------------------------
// enable and disable of interrupts
//
void globalInterruptEnable(PCIADA *pciada)
{
   WRITE_REGISTER_USHORT(pciada->pwIntCSR, ENABLE_PCIADA_IRQS);
}

void globalInterruptDisable(PCIADA *pciada)
{
   WRITE_REGISTER_USHORT(pciada->pwIntCSR, DISABLE_PCIADA_IRQS);
}

unsigned short globalInterruptEnabledStatus(PCIADA *pciada)
{
	unsigned short wStatus;

	wStatus = READ_REGISTER_USHORT(pciada->pwIntCSR);

	if ((wStatus & ENABLE_PCIADA_IRQS) == ENABLE_PCIADA_IRQS)
		return 1;

	return 0;
}
 
//------------------------------------------------------------------------
// determine which interrupt and evaluates status if appropriate
//
static ULONG evaluateIrqStatus(PCIADA *pciada)
{
	USHORT  wCntrl;
	KIRQL   oldIrql;
	ULONG   result = 0;

	// lock other users out
	KeAcquireSpinLock(&pciada->AccessLock, &oldIrql);

	wCntrl = READ_REGISTER_USHORT(pciada->pwCntrl);
    if (wCntrl & 0x100)   // pciada switched on ?
	{
		USHORT  wIntCSR = READ_REGISTER_USHORT(pciada->pwIntCSR);

		if ((wIntCSR & 0x68) == 0x68)
		{
			// it's the pci interrupt # 2

			// get current Cntrl - and clear interrupt
			wCntrl &= ~0x0100;  // disable
			WRITE_REGISTER_USHORT(pciada->pwCntrl, wCntrl);	
			wCntrl |= 0x0100;   // enable again
			WRITE_REGISTER_USHORT(pciada->pwCntrl, wCntrl);
			
			result = CONNECTION_TIMEOUT;
		}
		else // if it was a connection timeout - then it can't be a LAM interrupt
		{	
			if ((wIntCSR & 0x45) == 0x45)
			{
				result =  READ_REGISTER_ULONG(_DWORD_NAF(pciada->pvVirtIfr, 28, 2, 0));

				// clear pending interrupt - LAM-FF
				WRITE_REGISTER_USHORT(_WORD_NAF(pciada->pvVirtIfr, 28, 0, 16), 0);
			}
		}
	}
 
	// release the lock
	KeReleaseSpinLock(&pciada->AccessLock, oldIrql);

	return result;
}

//------------------------------------------------------------------------
// main interrupt handler function
//
BOOLEAN irq_service(PKINTERRUPT Interrupt, PVOID ServiceContext)
{
	PCIADA *pciada = (PCIADA *)ServiceContext;
	ULONG  dwIrqStatus;

	if (!(dwIrqStatus = evaluateIrqStatus(pciada))) 
	{
		KdPrint(("irq_service(0)\n"));
		return FALSE;
	}

	pciada->dwIrqStatus = dwIrqStatus;

	// fire custom deffered procedure call
	KeInsertQueueDpc(&pciada->kDPCobj, (PVOID)pciada, (PVOID)&pciada->dwIrqStatus);
    
	KdPrint(("irq_service(0x%08x)\n", pciada->dwIrqStatus));
	return TRUE;
}

//------------------------------------------------------------------------
//  translate interrupt resources for PCIADAs to neutral ones
//
NTSTATUS PCICC32TranslateInterrupt(PDEVICE_OBJECT device_Obj)
{
	int              i;
	NTSTATUS         result = STATUS_SUCCESS;
	int nPCIADAs = ((DEVICE_EXT*)(device_Obj->DeviceExtension))->nPCIADAs;
	PCIADA           *pciada;

    KdPrint(("TranslateInterrupt()\n"));

	for (i = 0; i < nPCIADAs; i++)
	{
		pciada = &((DEVICE_EXT*)(device_Obj->DeviceExtension))->pciada[i];

		KdPrint(("Bus:%d, IrqLine:%d\n", pciada->Bus, pciada->PCIDevice.u.type0.InterruptLine));

		if (pciada->PCIDevice.u.type0.InterruptLine)
			pciada->dwVirtVector = HalGetInterruptVector(PCIBus, pciada->Bus, 
				pciada->PCIDevice.u.type0.InterruptLine, 
					pciada->PCIDevice.u.type0.InterruptLine,
						&pciada->Irql, &pciada->Affinity);
	}

	return result;
}

//------------------------------------------------------------------------
//  connect service routines to all PCIADA interrupts
//
NTSTATUS PCICC32ConnectInterrupt(PDEVICE_OBJECT device_Obj)
{
	int              i;
	NTSTATUS         result = STATUS_SUCCESS;
	DEVICE_EXT       *pDevExt = (DEVICE_EXT*)device_Obj->DeviceExtension;
	int				 nPCIADAs = pDevExt->nPCIADAs;
	PCIADA           *pciada;
	KIRQL			 highestIrql = 0;

    KdPrint(("ConnectInterrupt()\n"));

	// check for highest Irql the device uses
	for (i = 0; i < nPCIADAs; i++)
	{
      pciada = &pDevExt->pciada[i];

	  if (pciada->Irql > highestIrql) 
		  highestIrql = pciada->Irql;
	}

	// connect the interrupts to service routines
	for (i = 0; i < nPCIADAs; i++)
	{
      pciada = &pDevExt->pciada[i];

	  pciada->InterruptObject = NULL;
	  
	  if (pciada->dwVirtVector)
		result = IoConnectInterrupt(&pciada->InterruptObject,
			irq_service,
				(PVOID)pciada, 
					NULL, 
						pciada->dwVirtVector,
							pciada->Irql,
								highestIrql,
									LevelSensitive,
										TRUE,
											pciada->Affinity,
												FALSE);

	  KdPrint(("pIrqObj:0x%08x, VirtVect:%d, Irql:%d, hIrql:%d, Aff:0x%08x, Status:0x%08x\n", 
          pciada->InterruptObject, 
				pciada->dwVirtVector, pciada->Irql, highestIrql, pciada->Affinity, result));

	  if (result != STATUS_SUCCESS) break;
	}

	return result;
}

//------------------------------------------------------------------------
//  dis-connect service routines to all PCIADA interrupts
//
NTSTATUS PCICC32DisConnectInterrupt(PDEVICE_OBJECT device_Obj)
{
	int              i;
	DEVICE_EXT       *pDevExt = (DEVICE_EXT*)device_Obj->DeviceExtension;
	int				 nPCIADAs = pDevExt->nPCIADAs;
	PCIADA           *pciada;

    KdPrint(("DisConnectInterrupt()\n"));

	// dis connect the interrupts to service routines
	for (i = 0; i < nPCIADAs; i++)
	{
		pciada = &pDevExt->pciada[i];

		KdPrint(("IrqObj:0x%08x\n", pciada->InterruptObject));

		if (pciada->InterruptObject)
		{
		  IoDisconnectInterrupt(pciada->InterruptObject);
		  pciada->InterruptObject = 0;
		}
	}

	return STATUS_SUCCESS;
}


