//-------------------------------------------------------------------------
//
// pcibios support functions - functions to access PCI bios directly
// (c) 1999 ARW Elektronik
//
// this source code is published under GPL (Open Source). You can use, redistrubute and
// modify it unless this header  is  not modified or deleted. No warranty is given that
// this software will work like expected.
// This product is not authorized for use as critical component in life support systems
// wihout the express written approval of ARW Elektronik Germany.
//
// Please announce changes and hints to ARW Elektronik
//
// what							who    when
// First parts born                                     AR     23.10.97
// get_interrupt_routing_options temp. removed          AR     28.05.99
//

#include <pcibios.h>
#include <calls.h>

#define PCI_ID 0x20494350                     // der PCI-check-code

//-------------------------------------------------------------------------
// interpretiert die fehler-codes und gibt einen string ohne CR-LF zurueck
//
char *interpret_error(int error_code)
{
  switch (error_code)
  {
    case PCI_SUCCESSFUL:          return "No error.";
    case PCI_FUNC_NOT_SUPPORTED:  return "Function not supported!";
    case PCI_BAD_VENDOR_ID:       return "Bad vendor ID!";
    case PCI_DEVICE_NOT_FOUND:    return "Device not found!";
    case PCI_BAD_REGISTER_NUMBER: return "Bad register number!";
    case PCI_SET_FAILED:          return "Set failed!";
    case PCI_BUFFER_TOO_SMALL:    return "Buffer too small!";

    default:                      return "Unknown error code!";
  }
}

//-------------------------------------------------------------------------
// fragt, ob ein PCI-BIOS vorhanden ist
//
int pcibios_present(int *major, int *minor, int *no_pci, int *mechanism)
{
  XREGS xregs;

  xregs.h.AH = PCI_FUNCTION_ID;
  xregs.h.AL = PCI_PCI_BIOS_PRESENT;

  pcibios(&xregs,&xregs);

  if ((!xregs.l.cflag) && (!xregs.h.AH) && (xregs.l.EDX == PCI_ID))
  {
    *major     = xregs.h.BH;
    *minor     = xregs.h.BL;
    *no_pci    = xregs.h.CL;
    *mechanism = xregs.h.AL;
  }
  else
  {
    *major     = 0;
    *minor     = 0;
    *no_pci    = 0;
    *mechanism = 0;
  }

  return xregs.h.AH;
}

//-------------------------------------------------------------------------
// fragt ob ein bestimmtes PCI-Device vorhanden ist - kann iterativ angewandt
// werden bis ein fehler zurueckkommt
//
int find_pci_device(int device_id, unsigned short vendor_id, int index,
			  int *bus_number, int *device_number, int *function_number)
{
  XREGS xregs;

  xregs.h.AH = PCI_FUNCTION_ID;
  xregs.h.AL = PCI_FIND_PCI_DEVICE;
  xregs.x.CX = device_id;
  xregs.x.DX = vendor_id;
  xregs.x.SI = index;

  pcibios(&xregs,&xregs);

  if ((!xregs.l.cflag) && (!xregs.h.AH))
  {
    *bus_number      = xregs.h.BH;
    *device_number   = (xregs.h.BL >> 3) & 0x1f;
    *function_number = xregs.h.BL & 0x7;
  }
  else
  {
    *bus_number      = 0;
    *device_number   = 0;
    *function_number = 0;
  }

  return xregs.h.AH;
}

//-------------------------------------------------------------------------
// sucht die PCI-devices einer bestimmten klasse - kann iterativ angewandt
// werden bis ein fehler auftritt.
//
int find_pci_class_code(unsigned long class_code, int index,
			  int *bus_number, int *device_number,
			    int *function_number)
{
  XREGS xregs;

  xregs.h.AH  = PCI_FUNCTION_ID;
  xregs.h.AL  = PCI_FIND_PCI_CLASS_CODE;
  xregs.l.ECX = class_code & 0xffffff;  // lower 3 bytes
  xregs.x.SI  = index;

  pcibios(&xregs,&xregs);

  if ((!xregs.l.cflag) && (!xregs.h.AH))
  {
    *bus_number      = xregs.h.BH;
    *device_number   = (xregs.h.BL >> 3) & 0x1f;
    *function_number = xregs.h.BL & 0x7;
  }
  else
  {
    *bus_number      = 0;
    *device_number   = 0;
    *function_number = 0;
  }

  return xregs.h.AH;
}

//-------------------------------------------------------------------------
// liest bytes, worte und langworte vom PCI-device
//
int read_configuration_byte(int bus_number, int device_number,
			      int function_number, int register_number,
				 char *result)
{
  XREGS xregs;

  xregs.h.AH  = PCI_FUNCTION_ID;
  xregs.h.AL  = PCI_READ_CONFIG_BYTE;
  xregs.h.BH  = bus_number;
  xregs.h.BL  = (device_number << 3) | (function_number & 0x07);
  xregs.x.DI  = register_number;

  pcibios(&xregs,&xregs);

  if ((!xregs.l.cflag) && (!xregs.h.AH))
    *result = xregs.h.CL;
  else
    *result = 0;

  return xregs.h.AH;
}

int read_configuration_word(int bus_number, int device_number,
			      int function_number, int register_number,
				 unsigned short *result)
{
  XREGS xregs;

  xregs.h.AH  = PCI_FUNCTION_ID;
  xregs.h.AL  = PCI_READ_CONFIG_WORD;
  xregs.h.BH  = bus_number;
  xregs.h.BL  = (device_number << 3) | (function_number & 0x07);
  xregs.x.DI  = register_number;

  pcibios(&xregs,&xregs);

  if ((!xregs.l.cflag) && (!xregs.h.AH))
    *result = xregs.x.CX;
  else
    *result = 0;

  return xregs.h.AH;
}


int read_configuration_dword(int bus_number, int device_number,
			      int function_number, int register_number,
				 unsigned long *result)
{
  XREGS xregs;

  xregs.h.AH  = PCI_FUNCTION_ID;
  xregs.h.AL  = PCI_READ_CONFIG_DWORD;
  xregs.h.BH  = bus_number;
  xregs.h.BL  = (device_number << 3) | (function_number & 0x07);
  xregs.x.DI  = register_number;

  pcibios(&xregs,&xregs);

  if ((!xregs.l.cflag) && (!xregs.h.AH))
    *result = xregs.l.ECX;
  else
    *result = 0;

  return xregs.h.AH;
}


//-------------------------------------------------------------------------
// schreibt bytes, worte und langworte vom PCI-device
//
int write_configuration_byte(int bus_number, int device_number,
			      int function_number, int register_number,
				 char value)
{
  XREGS xregs;

  xregs.h.AH  = PCI_FUNCTION_ID;
  xregs.h.AL  = PCI_WRITE_CONFIG_BYTE;
  xregs.h.BH  = bus_number;
  xregs.h.BL  = (device_number << 3) | (function_number & 0x07);
  xregs.x.DI  = register_number;
  xregs.h.CL  = value;

  pcibios(&xregs,&xregs);

  return xregs.h.AH;
}

int write_configuration_word(int bus_number, int device_number,
			      int function_number, int register_number,
				 unsigned short value)
{
  XREGS xregs;

  xregs.h.AH  = PCI_FUNCTION_ID;
  xregs.h.AL  = PCI_WRITE_CONFIG_WORD;
  xregs.h.BH  = bus_number;
  xregs.h.BL  = (device_number << 3) | (function_number & 0x07);
  xregs.x.DI  = register_number;
  xregs.x.CX  = value;

  pcibios(&xregs,&xregs);

  return xregs.h.AH;
}


int write_configuration_dword(int bus_number, int device_number,
			      int function_number, int register_number,
				 unsigned long value)
{
  XREGS xregs;

  xregs.h.AH  = PCI_FUNCTION_ID;
  xregs.h.AL  = PCI_WRITE_CONFIG_DWORD;
  xregs.h.BH  = bus_number;
  xregs.h.BL  = (device_number << 3) | (function_number & 0x07);
  xregs.x.DI  = register_number;
  xregs.l.ECX = value;

  pcibios(&xregs,&xregs);

  return xregs.h.AH;
}


//-------------------------------------------------------------------------
// erfragt die interrupt routing options - noch nicht fertig! -
//
/*
int get_interrupt_routing_options(unsigned short *irq_bitmap)
{
  XREGS xregs;

  xregs.h.AH  = PCI_FUNCTION_ID;
  xregs.h.AL  = PCI_GET_IRQ_ROUTING_OPTIONS;

  return PCI_FUNC_NOT_SUPPORTED;    // vorlaeufig
}
*/
