/*
 * Linux-Driver for the PCI-VMEbus Interface 
 * from ARW-Elektronik, D-69168 Wiesloch 
 * now by W-IE-NE-R, Plein Baus GmbH / Germany
 *
 * Author:
 * --------
 * Dirk Muehlenberg (IK3, Forschungszentrum Karlsruhe)
 *
 * Start Date: 8.2.99
 *
 * the driver files are:
 * ----------------------
 * arwvme.c arwvme.h arwvme_priv.h vic68.h vme.h
 * arwpci.c arwpci.h list.c list.h
 *
 * module name: arwvme.o
 *
 * Features:
 * ==========
 * MultiUser, MultiHardware, PnP (PCIbus), shared
 * interrupts (Linux version > 1.3.70)
 * VFS: read, write, seek, mmap, open, release, ioctl
 *
 * $Revision: 1.37 $ $Date: 1999/07/13 08:44:45 $
 *
 ! The plot for the correct mmap is:
 ! On an access to the VMEbus the nopage function will mmap the VMEwindow, 
 ! sets the correct window offset in the VIC68 and returns the physical 
 ! address of the VMEwindow. If this task looses the CPU (via schedule()), 
 ! the InvalidatePage() will unmap the VMEbus-window, so the next access will 
 ! hopefully generates a new page fault exception. If the current task will
 ! cross the page boundary, the new page is mapped and returned from nopage(). 
 *
 *
 ? There is a very important question about delivering
 ? a VMEbus error to the user (is the VMEbus an extension
 ? of the local hardware?):
 ? a) send SIGBUS,  
 ? b) printing a message to /var/log/messages and make the 
 ?    entries to the structure ArwvmeSysErrInfo. 
 ? c) kernel panic
 ? For now I print the message (b), make the entry and send
 ? SIGBUS (a), but the other solution (c) is realized in 
 ? the VMEbus driver for the VME-Alpha 224/4, the authors of 
 ? the device driver let the kernel panic on a VMEbus write access
 ? error (perhaps there is no other possibility due to 
 ? the design of the embedded Alpha). I think sending
 ? a SIGBUS (a) like a unprocessed page error is a good 
 ? compromise for the user, a kernel panic is unacceptable.
 ? I have included the SIGBUS support (-DSIGBUSON), but the
 ? user have to remember to register a handler for the SIGBUS 
 ? else the program will be terminated. With the handler supporting 
 ? infos about the nature of the error the
 ? user can differ between a VMEbus Error and other BusErrors
 ? via ArwvmeGetSysError()->reason, in the cases of PCIADA_TO, 
 ? ACFAIL or SYSFAIL the process should exit(). 
 ? If you have not registered and you are running a debugger, 
 ? you get a sigbus in the code section behind the faulty access!
 *
 ! Errors like SYSFAIL, ACFAIL are now flagged via SIGBUS.
 ! The occurence of VMEbus error is handled async via
 ! irq handler, but sync in read(), write() to stop
 ! the misbehaving transmission returning the so far
 ! correctly transmitted number of bytes. (look also 
 ! previous section)
 *
 * Shared interrupts:
 * ------------------
 * you have to pass the dev_id to request_irq.
 * the dev_id is (vendor_id << 16) | device_id,
 * 'cos the kernel calls ALL handlers sharing this
 * interrupt but with only this VENDOR-DEVICE_ID 
 * pair.
 *
 * Resident Set Size:
 * there is a bug in the linux kernel:
 * the rss entry in mm_struct is incremented on 
 * every page fault, but for reserved pages never
 * decreased, resulting in a continous growing of
 * this value (the max is INT_MAX). If you encounter
 * ps l the rss value is shown and perhaps very large.
 * But there is no resource connected to this value
 * in case of reserved pages, so I decrease it on every
 * page fault referring to my VME window.
 *  
 * Acknowledgements:
 * ------------------
 * The author thanks A.Rausch for loaning him the interface and for
 * the support in hardware debugging during the first phase of the
 * driver development.
 * Further thanks for testing with other Vme hardware to:
 * - Lueder v. Horn (vhorn@wiener-d.com)
 * - Klaus Hitschler (Klaus.Hitschler@gmx.de)
 * - Juergen Hannappel (hannappe@lisa2.physik.uni-bonn.de)
 *
 */

#include <linux/unistd.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/wrapper.h>
#include <linux/interrupt.h>

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/mman.h>

#include <linux/arwvme.h>
#include <linux/vme.h>

#include "arwvme_priv.h"

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,4)
#include <asm/uaccess.h>
#include <linux/poll.h>
#define put_fs_long(x,ptr) put_user(x,(unsigned long *)ptr) 
#endif

#include "arwpci.c"
#include "list.c"

extern void console_print(const char *);

#ifdef DEBUG
#define ArwvmePrint(x) ArwvmeConsolePrint x
#else
#define ArwvmePrint(x) {}
#endif

/* ----- Global declarations ----------------------------------------------- */
static List *pci_device_header = NULL;
static List *hardwareHandle = NULL;

/* fast access to HardwareHandle */
#if LINUX_VERSION_CODE < LINUX_NO_SHIRQ 
static ArwvmeHardwareHandle *irq2hh[NR_IRQS];
#endif
static ArwvmeHardwareHandle *modnr2hh[0x10];

void cleanup_module(void);
int init_module(void);
void arwvme_init(void);
void arwvme_setup(char *str, int *ints);
int bad_user_access_length(void);

/* ----- private functions ------------------------------------------------- */
static ArwvmeUserHandle *ArwvmeGetUserHandleFromAM(unsigned char am);
static ArwvmeUserHandle *ArwvmeGetUserHandleFromHardware(ArwvmeHardwareHandle *hh);
static ArwvmeHardwareHandle *ArwvmeGetHardwareHandleFromIndex(int index);
static ArwvmeHardwareHandle *ArwvmeGetHardwareHandleFromModuleNumber(int mn);
static int Arwvme_check_connection(ArwvmeHardwareHandle *hh);
static int ArwvmeRequestIOPorts(ArwvmeHardwareHandle *hh);
static List *ArwvmeSetupHardwareHandle(List *pch);
static void ArwvmeFreeHardwareHandle(ArwvmeHardwareHandle *hh);
static void ArwvmeFreeUserHandle(ArwvmeUserHandle *uh);
static ArwvmeIntAck *ArwvmeAllocIntAckVector(void);
static void Arwvme_init(ArwvmeHardwareHandle *hh);
static void Arwvme_deinit(ArwvmeHardwareHandle *hh);
static int ArwvmeSearchPIDInUserList(void *d1, void *d2);
static ArwvmeUserHandle *ArwvmeGetUserHandleFromPID(ArwvmeHardwareHandle *hh,
                                                    pid_t pid);
static ArwvmeUserHandle *ArwvmeGetUserHandleFromCurrent(ArwvmeHardwareHandle *hh);
static void Arwvme_reset(ArwvmeHardwareHandle *hh);
static unsigned char Arwvme_build_am(unsigned short addr_type, unsigned short acc_type);
static int Arwvme_set_am(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh, ArwvmeADA *ada);
static void *Arwvme_build_address(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh, unsigned int addr);
static void Arwvme_switch_context(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh);
static unsigned long arwvme_vma_nopage(struct vm_area_struct * area, unsigned long address, int no_share);
static void arwvme_vma_open(struct vm_area_struct * area);
static void arwvme_vma_close(struct vm_area_struct * area);
static int Arwvme_enable_interrupt(ArwvmeHardwareHandle *hh);
static void arwvme_irq_handler(int irq, void *idontknow, struct pt_regs *ptr);
static void Arwvme_free_irq(ArwvmeHardwareHandle *hh, int irq);
static void ArwvmeHandleSystemInterrupt(ArwvmeHardwareHandle *hh);
static int ArwvmeAddInterruptInfo(ArwvmeHardwareHandle *hh, unsigned long tl);
static void ArwvmeRoraMethod(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh);
static void ArwvmeResetSysErrInfo(ArwvmeUserHandle *uh);
static void ArwvmeSetSysErrInfo(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh);
static ArwvmeUserHandle *ArwvmeInitUserHandle(void);
static void ArwvmeInvalidatePage(ArwvmeHardwareHandle *);
static void ArwvmeConsolePrint(const char *fmt, ...);
static void ArwvmePrintPageSetup(unsigned long);
static int ArwvmeGetUserValue(int cmd, void *to, void *from);
static int ArwvmePutKernelValue(int cmd, void *to, void *from);
static void ArwvmeKickAllUsers(ArwvmeHardwareHandle *hh);

static void ArwvmeWriteA_D8(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                            off_t f_pos, int data);
static void ArwvmeWriteA_D16(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                            off_t f_pos, int data);
static void ArwvmeWriteA_D32(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                            off_t f_pos, int data);
static int ArwvmeReadA_D8(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                          off_t f_pos);
static int ArwvmeReadA_D16(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                          off_t f_pos);
static int ArwvmeReadA_D32(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                          off_t f_pos);

/* ----- Virtual File System ----------------------------------------------- */
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,4)
static loff_t arwvme_lseek( struct inode * inode, struct file * file,
                                 off_t offset, int origin );
#else
static int arwvme_lseek( struct inode * inode, struct file * file,
                                 off_t offset, int origin );
#endif
static ssize_t arwvme_read ( struct inode *inode, struct file *file,
                                 char *buf,  int count );
static ssize_t arwvme_write ( struct inode *inode, struct file *file,
                                 const char *buf, int count );
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,4)
static unsigned int arwvme_poll(struct file *file, poll_table *wait );
#else
static int arwvme_select(struct inode *inode, struct file *file,
                         int sel_type, select_table *wait );
#endif
static int arwvme_ioctl ( struct inode *inode, struct file *file,
                                 unsigned int cmd, unsigned long arg );
static int arwvme_mmap ( struct inode *inode, struct file *file,
                                 struct vm_area_struct * );
static int arwvme_open ( struct inode *inode, struct file *file );
static void arwvme_release ( struct inode *inode, struct file *file );

/* needed by mmap to catch the page fault */
static struct vm_operations_struct arwvme_vmops = {
  arwvme_vma_open,   /* open */
  arwvme_vma_close,  /* close */
  NULL,              /* unmap */
  NULL,              /* protect */
  NULL,              /* sync */
  NULL,              /* advise */
  arwvme_vma_nopage, /* nopage */
  NULL,              /* wppage */
  NULL,              /* swapout */
  NULL,              /* swapin */
};

/* --- function tables used by arwvme_read(), arwvme_write() --- */

static int (*ArwvmeReadVTBL[])(ArwvmeHardwareHandle *, ArwvmeUserHandle *, 
                               off_t) = 
{
  ArwvmeReadA_D8,
  ArwvmeReadA_D16,
  ArwvmeReadA_D32
};

static void (*ArwvmeWriteVTBL[])(ArwvmeHardwareHandle *, ArwvmeUserHandle *,
                                 off_t, int) = 
{  
  ArwvmeWriteA_D8,
  ArwvmeWriteA_D16,
  ArwvmeWriteA_D32
};



/* ------ DM 23.2.99 ------------------------------------------------------- */
static void ArwvmeConsolePrint(const char *fmt, ...)
{
  static char buf[1024];
  va_list ap;
  va_start(ap,fmt);
  vsprintf(buf, fmt, ap);
  va_end(ap);
  console_print(buf);
}

/* ------ DM 18.2.99 ------------------------------------------------------- */

/* 
 * this is to use the macros __put_user() and __get_user()
 */

int bad_user_access_length(void)
{  
  printk("ARWVME::bad_user_access_length(): pid %d\n", current->pid);
/* 
 * originally in kernel/panic.c:
 * panic("ARWVME::bad_user_access_length(): pid %d\n", current->pid); 
 */ 
  return 0;
}

/* -------- DM 1.6.99 ------------------------------------------------------ */
static void ArwvmeKickAllUsers(ArwvmeHardwareHandle *hh)
/*
 * sends SIGBUS to all users
 * if a user will caught, he
 * will find the reason for the
 * kick in ArwvmeSysErrInfo.reason
 * else he dies.
 */
{
  Node *node;
  ArwvmeUserHandle *uh;

  LOCK_HH(hh);
  node = getFirstNode(hh->mm_user);
  while (node) {
    uh = (ArwvmeUserHandle *)node->data;
    kill_proc(uh->pid, SIGBUS, 0);
    node = node->succ;
  }
  RELEASE_HH(hh);
}

/* -------- DM 9.2.99 ------------------------------------------------------ */
static void ArwvmeUpdateUserHandleFromAM(ArwvmeUserHandle *uh, unsigned char am)
{
  int i,diff;
  static int offs[] = {0x01, 0x09, 0x29, 0x39, 0x40};
  static int addr[] = {ARWVME_A64, ARWVME_A32, ARWVME_A16, ARWVME_A24};
  static int acc[] = {ARWVME_DATA, ARWVME_PROG, ARWVME_BLOCK};

  if (uh->vmed.am == am) return;
  uh->pid = current->pid;
  uh->vmed.am = am;
  uh->vmed.ada.addr_type = 0; 
  uh->vmed.ada.data_type = 0; 
  uh->vmed.ada.acc_type = 0; 
  if (am > 0xf && am < 0x20) return;
  for (i=0;i<NUMBER(offs)-1;i++) {
    if (am >= offs[i] && am < offs[i+1]) {
      uh->vmed.ada.addr_type = addr[i]; 
      diff = am - offs[i];
      if (diff > 3)  {
        uh->vmed.ada.acc_type = ARWVME_SUP;
      } else {
        uh->vmed.ada.acc_type = ARWVME_NONPRIV;
      }
      if (i == 2) return;
      if (diff > 3)  {
        uh->vmed.ada.acc_type |= acc[diff - 0x05]; 
      } else {
        uh->vmed.ada.acc_type |= acc[diff];
      }
      return;
    } 
  }
}

/* -------- DM 22.4.99 ----------------------------------------------------- */
static void ArwvmeSetSysErrInfo(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh)
{
  uh->sys_err.module_number = hh->module_number;
  uh->sys_err.vme_addr = hh->current_uh->vmed.vme_addr;
  uh->sys_err.jiffies = jiffies;
  uh->sys_err.reason = hh->int_vec;
}

/* -------- DM 22.4.99 ----------------------------------------------------- */
static void ArwvmeResetSysErrInfo(ArwvmeUserHandle *uh)
{
  uh->sys_err.jiffies = 0;
  uh->sys_err.module_number = -1;
  uh->sys_err.vme_addr = 0;
  uh->sys_err.reason = 0;
}

/* -------- DM 27.5.99 ----------------------------------------------------- */
static ArwvmeIntAck *ArwvmeAllocIntAckVector(void)
{
  int i;
  ArwvmeIntAck *ret = vmalloc(sizeof(ArwvmeIntAck)*0x100);
  if (ret) {
    memset(ret,0,sizeof(ArwvmeIntAck)*0x100);
    for (i=0;i<0x100;i++)
      ret[i].int_vec = -1;
  }
  return ret;
}

/* -------- DM 12.2.99 ----------------------------------------------------- */
static ArwvmeUserHandle * ArwvmeInitUserHandle(void)
{
  ArwvmeUserHandle *uh = ANEW(ArwvmeUserHandle);
  if (!uh) return NULL; 

  ARWVME_VMED_INIT(uh->vmed);

  uh->pid = current->pid;
  uh->s_pid = 0;
  uh->fpos_tied = TRUE;
  uh->pendingInts = newList();
  uh->intInInts = FALSE;
  ArwvmeResetSysErrInfo(uh);
  uh->int_ack = ArwvmeAllocIntAckVector();

  ARWVME_MMAPD_INIT(uh->mmapd);

  return uh;
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static ArwvmeUserHandle *ArwvmeGetUserHandleFromAM(unsigned char am)
{
  ArwvmeUserHandle *user = ArwvmeInitUserHandle();
  if (!user) return NULL;
  ArwvmeUpdateUserHandleFromAM(user, am);
  return user;
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static ArwvmeUserHandle *ArwvmeGetUserHandleFromHardware(ArwvmeHardwareHandle *hh)
{
  ArwvmeUserHandle *uh = ArwvmeGetUserHandleFromAM(ARWVME_RD_ADDR_MODIFIER(hh->vic_base));
  if (!uh) return NULL;

  uh->vmed.vme_addr = ARWVME_RD_VMEADDR_REG_L(hh->mem_base);
  uh->vmed.arbiter = readb(hh->vic_base+VIC68_ARCR);
  uh->vmed.vmebus_to = readb(hh->vic_base+VIC68_TTR);

  return uh;
}

static void ArwvmeUpdateUserHandleFromHardware(ArwvmeHardwareHandle *hh,
                                               ArwvmeUserHandle *uh)
{
  ArwvmeUpdateUserHandleFromAM(uh,ARWVME_RD_ADDR_MODIFIER(hh->vic_base));
  uh->vmed.vme_addr = ARWVME_RD_VMEADDR_REG_L(hh->mem_base);
}

/* -------- DM 9.2.99 ------------------------------------------------------ */
static ArwvmeHardwareHandle *ArwvmeGetHardwareHandleFromModuleNumber(int mn)
{
  Node *n;
  ArwvmeHardwareHandle *hh = NULL;
  n = getFirstNode(hardwareHandle);
  while (n) {
    hh = (ArwvmeHardwareHandle *)(n->data);
    if (hh->module_number == mn) return hh; 
    n = n->succ;
  }  
  return NULL;
}

/* -------- DM 17.2.99 ----------------------------------------------------- */
static ArwvmeHardwareHandle *ArwvmeGetHardwareHandleFromIndex(int index)
{
  Node *n;
  ArwvmeHardwareHandle *hh = NULL;
  n = getFirstNode(hardwareHandle);
  while (n) {
    hh = (ArwvmeHardwareHandle *)(n->data);
    if (hh->index == index) return hh; 
    n = n->succ;
  }  
  return NULL;
}

/* -------- DM 12.2.99 ----------------------------------------------------- */
void ArwvmeRoraMethod(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh)
{
  ArwvmeIntAck *ia = uh->int_ack+hh->int_vec;
  int data_type;
  ArwvmeADA ad; 
  unsigned char old_am;
  unsigned int lcr_base = ARWVME_LCR_BASE(hh);

  old_am = ARWVME_RD_ADDR_MODIFIER(hh->vic_base);

/*
 * prepare the correct am
 * for the rora register
 */
  ad.addr_type = ia->acc_mask & ARWVME_A_MASK;
  ad.data_type = ia->acc_mask & ARWVME_D_MASK;
  ad.acc_type = ia->acc_mask & ( ARWVME_ACC_MASK | ARWVME_PRIV_MASK );

  Arwvme_set_am(hh, uh, &ad);
  data_type = ((ad.data_type & ARWVME_D_MASK)>>4)-1;

  ArwvmeWriteVTBL[data_type](hh, uh, ia->rora_reg, 1);
  if (ARWVME_GET_BUS_ERR(hh->vic_base)) {
    ARWVME_CLR_INT2(lcr_base);
    ArwvmeReadVTBL[data_type](hh, uh, ia->rora_reg);
    if (ARWVME_GET_BUS_ERR(hh->vic_base)) {
      ARWVME_CLR_INT2(lcr_base);
      printk("atvme_rora_method(): error occured during register access!\n");
    }
  }
  ARWVME_WR_ADDR_MODIFIER(hh->vic_base, old_am);
}

/* -------- DM 19.2.99 ----------------------------------------------------- */
static void ArwvmeHandleSystemInterrupt(ArwvmeHardwareHandle *hh)
/*
 * handles unusual conditions:
 * PCIADA_TO, ACFAIL, SYSFAIL leads
 * to a SIGBUS, which the user have to
 * catch to survive!
 */ 
{
  static ArwvmeSysErrInfo sei;

  if (hh->int_vec > ARWVME_MAX_VECTOR) return;

  sei.module_number = hh->module_number;
  sei.vme_addr = hh->current_uh->vmed.vme_addr;
  sei.jiffies = jiffies;
  sei.reason = hh->int_vec;
  memcpy(&hh->current_uh->sys_err, &sei, sizeof(ArwvmeSysErrInfo));

  switch (hh->int_vec) {
    case ARWVME_PCIADA_TO     : 
      printk("ArwvmeHandleSystemInterrupt(): check the cable and the VME-crate for PCIADA/VMEMM %d!\n", hh->module_number);
      ArwvmeKickAllUsers(hh);
    break;
    case ARWVME_RESET_BUTTON  :
      printk("ArwvmeHandleSystemInterrupt(): reset button pressed on PCIADA/VMEMM %d!\n", hh->module_number);
    break;
    case ARWVME_VMEBUS_TO     :
      if (!hh->in_rw) {
        ARWVME_CLR_BESR(hh->vic_base);
      }
#ifdef DEBUG
      printk("ArwvmeHandleSystemInterrupt(): error occured during VMEbus access on PCIADA/VMEMM %d from user %d!\n", hh->module_number, hh->current_uh->pid);
#endif
#if SIGBUSON        
      kill_proc(current->pid, SIGBUS, 0);
#endif
    break;
    case ARWVME_ACFAIL        :
      printk("ArwvmeHandleSystemInterrupt(): ACFAIL on PCIADA/VMEMM %d!\n", hh->module_number);
      ArwvmeKickAllUsers(hh);
    break;
    case ARWVME_SYSFAIL       :
      printk("ArwvmeHandleSystemInterrupt(): SYSFAIL on PCIADA/VMEMM %d!\n", hh->module_number);
      ArwvmeKickAllUsers(hh);
    break;
    default:
      printk("ArwvmeHandleSystemInterrupt(): this hardware interrupt 0x%x occuring on PCIADA/VMEMM %d is not handled!\n", hh->int_vec, hh->module_number); 
  }
}

/* -------- DM 12.2.99 ----------------------------------------------------- */
static int ArwvmeAddInterruptInfo(ArwvmeHardwareHandle *hh, unsigned long int_tl)
/*
 * scroll through all users of this VMEMM
 * to check if they want this irq:
 * 1. do rora, if needed
 * 2. add IntInfo to the user queues 
 */ 
{
  ArwvmeUserHandle *uh;
  ArwvmeIntAck *user_int_ack;
  int max;
  Node *uhn;
  static char rora_bits[256];

  if (hh->int_vec <= ARWVME_MAX_VECTOR) {
/* this was a PCIADA or VMEMM initiated interrupt */ 
    ArwvmeHandleSystemInterrupt(hh); 
    return 0; 
  }

  memset(rora_bits,0,sizeof(rora_bits));
  uhn = getFirstNode(hh->mm_user);
  while (uhn) { 
    uh = (ArwvmeUserHandle *)uhn->data;
    if (!uh->intInInts) {
      uhn = uhn->succ;
      continue;
    }
    if (hh->int_vec == uh->int_ack[hh->int_vec].int_vec) {
      user_int_ack = ANEW(ArwvmeIntAck);
      if (!user_int_ack) {
        ArwvmeConsolePrint("ArwvmeAddInterruptInfo(): no memory for struct ArwvmeIntAck!\n");
        return -ENOMEM;
      }
      if (uh->int_ack[hh->int_vec].ack_type == ArwvmeRoraByHandler && !rora_bits[1<<hh->int_vec]) {
        ArwvmeRoraMethod(hh, uh);
        rora_bits[1<<hh->int_vec] = 1;
      }
      memcpy(user_int_ack, uh->int_ack+hh->int_vec,sizeof(ArwvmeIntAck));
      user_int_ack->int_tl = int_tl;
      if ((max = getNumOfNodesInList(uh->pendingInts)) > ARWVME_MAX_INTS_QUEUED) {
#ifdef DISABLE_INTS_ON_NEGLECT
        deleteList(uh->pendingInts,NULL);
        uh->pendingInts = NULL;
        uh->intInInts = FALSE; 
        ArwvmeConsolePrint("ArwvmeAddInterruptInfo(): max. %d interrupts pending in queue for pid %d. Removing queue now!\n", max, uh->pid);
        FREE(user_int_ack);
        uhn = uhn->succ;
        continue;
#else
        resetList(uh->pendingInts,NULL);
        ArwvmeConsolePrint("ArwvmeAddInterruptInfo(): max. %d interrupts pending in queue for pid %d. Erazing interrupts now!\n", max, uh->pid);
#endif
      }
      addTail(uh->pendingInts, user_int_ack);
      if (uh->s_pid > 0) 
        kill_proc(uh->s_pid, SIGUSR1, 0);
    } else {
      ArwvmeConsolePrint("ArwvmeAddInterruptInfo(): int_vec %d not registered for this handler!\n",hh->int_vec);
    }
    uhn = uhn->succ;
  }
  hh->int_vec = 0;
  wake_up_interruptible(&hh->wait_queue);
  return 0;
}

/* -------- DM 12.2.99 ----------------------------------------------------- */
static void arwvme_irq_handler(int irq, void *dev_id, struct pt_regs *ptr)
{
  ArwvmeHardwareHandle *hh;
  unsigned long flags;
  unsigned int lcr_base;

#if LINUX_VERSION_CODE < LINUX_NO_SHIRQ
  hh = irq2hh[irq];
  if (!hh) {
    printk("arwvme_irq_handler(): no device found for irq %d!\n", irq);
    return;
  }
#else
/*
 * Ok, we are sharing.
 */ 
  hh = (ArwvmeHardwareHandle *)dev_id;
  if (!VALIDATE_HH(hh)) {
    return; 
  }
#endif

  save_flags(flags);
  cli();

  lcr_base = ARWVME_LCR_BASE(hh);

  if (!ARWVME_TST_INT(lcr_base)) {
#if LINUX_VERSION_CODE < LINUX_NO_SHIRQ
    printk("ARWVME: spurious interrupt for irq %d!\n",irq); 
#endif
    restore_flags(flags);
    return;
  }

  if (ARWVME_TST_PCIADA_INT(lcr_base)) {
    if (!ARWVME_VMEMM_OK(lcr_base)) {
      ARWVME_CLR_INT2(lcr_base);
      ARWVME_CLR_BESR(hh->vic_base);
      Arwvme_free_irq(hh, irq);
      printk("arwvme_irq_handler(): can't access VMEMM module %d!\n",
             hh->module_number);
      restore_flags(flags);
      return;
    }
  }

  hh->int_vec = ARWVME_READ_INT_VEC(hh->mem_base);
  if (hh->int_vec == 0xFF || hh->int_vec == 0) {
    printk("arwvme_irq_handler(): no vector for interrupt on module %d!\n",
                 hh->module_number);
    ARWVME_CLR_INT2(lcr_base);
    ARWVME_CLR_BESR(hh->vic_base);
    Arwvme_free_irq(hh, irq);
    restore_flags(flags);
    return;
  }

/* no bottom half here, 'cos not much to do */
  ArwvmeAddInterruptInfo(hh, jiffies);

  restore_flags(flags);
}

/* -------- DM 12.2.99 ----------------------------------------------------- */
static void Arwvme_free_irq(ArwvmeHardwareHandle *hh, int irq)
{
  if (irq > 0) {
    disable_irq(irq);
#if LINUX_VERSION_CODE < LINUX_NO_SHIRQ
    irq2hh[irq] = NULL;
    free_irq(irq, NULL);
#else
    if (!hh) return;
    free_irq(irq, hh);
#endif
  }
  ArwvmePrint(("Arwvme_free_irq(): freeing irq %d on module %d\n",
                irq, hh->module_number));
  hh->irq = 0;
}

/* -------- DM 12.2.99 ----------------------------------------------------- */
static int Arwvme_enable_interrupt(ArwvmeHardwareHandle *hh)
{
  int ret; 

  if (hh->irq) {
/* this irq is still in use for this device */
    return 0;
  }

  ArwvmePrint(("Arwvme_enable_interrupt() for index %d ...\n", hh->index));

/* 
 * Are shared ints supported?
 */
#if LINUX_VERSION_CODE < LINUX_NO_SHIRQ
  ret = request_irq(hh->pci_conf->irq, arwvme_irq_handler, 
                        SA_INTERRUPT, ARWVME_NAME, NULL);
  ArwvmePrint(("request_irq(): no sharing\n"));
#else
  ret = request_irq(hh->pci_conf->irq, arwvme_irq_handler, 
                        SA_INTERRUPT | SA_SHIRQ, ARWVME_NAME, hh);
  ArwvmePrint(("request_irq(): hh = %p\n",hh));
#endif
  if (ret) {
    switch (-ret) {
      case EINVAL:
        printk("Arwvme_enable_interrupt(): invalid irq line %d!\n", hh->pci_conf->irq);
        return ret;
      case ENOMEM: 
        printk("Arwvme_enable_interrupt(): no memory for irq handler (irq %d)!\n",hh->pci_conf->irq);
        return ret;
      case EBUSY:
        printk("Arwvme_enable_interrupt(): irq %d already busy!\n", hh->pci_conf->irq); 
        return ret; 
      default:
        printk("Arwvme_enable_interrupt(): could not install irq handler for irq %d!\n",hh->pci_conf->irq);
/* this irq is not available for this board! */
        hh->pci_conf->irq = 0;
    }  
  } else {
    hh->irq = hh->pci_conf->irq;   
#if LINUX_VERSION_CODE < LINUX_NO_SHIRQ
    irq2hh[hh->pci_conf->irq] = hh;
#endif
    printk("Arwvme_enable_interrupt(): Handler for irq %d successfully installed.\n",hh->pci_conf->irq);

    ARWVME_EN_INT(ARWVME_LCR_BASE(hh));
/*
 * ARWVME_EN_INT() enables 2 int sources: PCIADA, VMEMM
 * PCIADA interrupts:
 *  Timeout on VMEbus (24us)
 *  data cable not connected
 *  crate has no power
 *  PCI-VME interface has no power or is not enabled
 * VMEMM interrupts:
 *  the 7 VMEbus levels
 *  mailbox int of VIC68a
 *  interrupt based on succesful interrupt generation
 *  SYSFAIL, ACFAIL 
 *  Bus timeout
 *  Arbiter timeout
 *  VIC68a timer interrupts
 *  reset button pressed
 * default vectorisation of the driver (0x0..0x3f):
 * (s. Arwvme_init())
 *  1 = PCIADA timeout                  (a)
 *  2 = clock tick interrupt generator  (na)
 *  6 = reset button                    (a)
 *  7 = VMEbus timeout (BERR)           (a)
 *  8 = interprocess global switch #0   (na)
 *  9 = interprocess global switch #1   (na)
 * 10 = interprocess global switch #2   (na)
 * 11 = interprocess global switch #3   (na)
 * 12 = interprocess module switch #0   (na)
 * 13 = interprocess module switch #1   (na)
 * 14 = interprocess module switch #2   (na)
 * 15 = interprocess module switch #3   (na)
 * 16 = ACFAIL asserted                 (na)
 * 17 = Write Post failed               (na) 
 * 18 = Arbitration timeout             (na)
 * 19 = SYSFAIL asserted                (na)
 * 20 = VMEbus interrupter acknowledge  (na)
 */
    enable_irq(hh->pci_conf->irq);
  }
  return ret;
}

static char *ArwvmePrintUserHandle(void *v)
{
  static char buf[256];
  ArwvmeUserHandle *uh = (ArwvmeUserHandle *)v;
  sprintf(buf, "PID: %d ", uh->pid);
  return buf;
}

/*
 * VFS entries
 */
/* -------- DM 9.2.99 ------------------------------------------------------ */
static int arwvme_open(struct inode *inode, struct file *file)
/*
 * give the minor id for board access:
 * arwvme0, arwvme1, ...
 */
{
  ArwvmeHardwareHandle *hh;

  ArwvmePrint(("arwvme_open() for user %d ...\n", current->pid));

  hh = modnr2hh[MINOR(inode->i_rdev)];
  if (!hh) {
    printk("arwvme_open(): no such device!\n");
    return -ENODEV;
  }

  ArwvmePrint(("... MOD_IN_USE %ld ...\n",MOD_IN_USE));

  LOCK_HH(hh);
/*
 * no user aboard -> reset 
 */  
  if (getNumOfNodesInList(hh->mm_user) < 1) {
    Arwvme_reset(hh);
    Arwvme_enable_interrupt(hh);
  } else {
    if (ArwvmeGetUserHandleFromCurrent(hh)) {
/*
 * Ooops, this user is still holding this crate
 */
      RELEASE_HH(hh);
      return -EBUSY;
    }
  }

  MOD_INC_USE_COUNT;

  hh->current_uh = ArwvmeGetUserHandleFromHardware(hh);
  /* the user has not chosen any */
  /* hardware configuration yet  */
  addTail(hh->mm_user, hh->current_uh);
  ArwvmePrint(("... adding user pid %d.\n",current->pid));
  file->private_data =  hh->current_uh;
  RELEASE_HH(hh);

  return 0;
}

/* -------- DM 9.2.99 ------------------------------------------------------ */
static void arwvme_release(struct inode *inode, struct file *file)
{
  ArwvmeHardwareHandle *hh = modnr2hh[MINOR(inode->i_rdev)];
  ArwvmeUserHandle *uh;
  Node *node;

  if (!hh) {
    printk("arwvme_release(): no such device!\n");
    return;
  }

  uh = (ArwvmeUserHandle *)file->private_data;
  if (!uh) {
    printk("arwvme_release(): no UserHandle!\n");
    return;
  }

  LOCK_HH(hh);
  node = getFirstNode(hh->mm_user);
  while (node) {
    if (node->data == uh) {
      hh->current_uh = NULL;
#if 1
      printk("arwvme_release(): pg_cnt = %d, rss %ld for pid %d\n", 
             uh->mmapd.pg_cnt, current->mm->rss, uh->pid);
#endif
      delNodeInList(hh->mm_user, node);
      ArwvmeFreeUserHandle(uh);
      file->private_data = NULL;
      MOD_DEC_USE_COUNT;
      node = NULL;
    } else {
      node = node->succ;
    }
  }

  if (getNumOfNodesInList(hh->mm_user) < 1) {
    Arwvme_deinit(hh);
    if (hh->irq) Arwvme_free_irq(hh,hh->irq);
#ifdef DEBUG
    printk("... all users released on module %d\n",hh->module_number);
#endif
  }
  RELEASE_HH(hh);
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,4)
static loff_t arwvme_lseek(struct inode * inode, struct file * file,
                                              off_t offset, int origin)
#else
static int arwvme_lseek(struct inode * inode, struct file * file,
                                              off_t offset, int origin)
#endif
{
  ArwvmeHardwareHandle *hh = modnr2hh[MINOR(inode->i_rdev)];
  ArwvmeUserHandle *uh;

  ArwvmePrint(("arwvme_lseek() for user %d ...\n", current->pid));

  if (!hh) {
    printk("arwvme_lseek(): no such device!\n");
    return -ENODEV;
  }

  uh = (ArwvmeUserHandle *)file->private_data;
  if (!uh) {
    return -EBADF;
  }

  if (!uh->vmed.ada.data_type) {
    printk("arwvme_lseek(): user %d has not chosen data type yet!\n", current->pid);
    return -EIO;
  }

  switch (origin) {
    int dtype;
    case 0:
      dtype = (1 << ARWVME_DATA_TYPE_INDEX(uh->vmed.ada.data_type))-1;
      if (offset & dtype) return -EFAULT;
      Arwvme_build_address(hh, uh, offset);
      file->f_pos = offset;
      ArwvmePrint(("... seeking f_pos 0x%lx.\n",(long int)file->f_pos)); 
    break;
    default: return -EINVAL;
  }
  return file->f_pos;
}

/* mmap stuff */
/* -------- DM 1.3.99 ------------------------------------------------------ */
static void ArwvmeInvalidatePage(ArwvmeHardwareHandle *hh)
/*
 * This routine is called from schedule() 
 * before the task switch takes place: 
 * current is the task, which will be moved
 * to the end of the queue. 
 * This routine will mark the VMEwindow as unused
 * and the end of the scheduling of the current_uh.
 * This routine is called directly from arwvme_read(),
 * arwvme_write() to avoid mmapped access to a wrong
 * page.
 */
{
  ArwvmeUserHandle *uh = hh->current_uh;
  unsigned long ret;

  if (!uh) return;

/* using no mmap at all or already unmapped */
  if (!uh->mmapd.reserved_page || !uh->mmapd.page_status) return; 

  ret = do_mmap(uh->mmapd.file, uh->mmapd.reserved_page, uh->mmapd.size, 
          PROT_READ|PROT_WRITE, MAP_FIXED|MAP_SHARED, hh->vme_base_p); 
  if (ret < 0) {
    printk("ArwvmeInvalidatePage() after do_mmap(): ret = %ld\n",ret);
  }
  uh->mmapd.page_status = 0;
  uh->mmapd.pg_cnt++;
}

/* -------- DM 2.3.99 ------------------------------------------------------ */
void arwvme_vma_open(struct vm_area_struct *vma)
{
  ArwvmeHardwareHandle *hh = modnr2hh[MINOR(vma->vm_inode->i_rdev)];
  MOD_INC_USE_COUNT;
  if (hh) {
    mem_map_inc_count(MAP_NR(hh->vme_base_p));
  }
}

void arwvme_vma_close(struct vm_area_struct *vma)
{
  ArwvmeHardwareHandle *hh = modnr2hh[MINOR(vma->vm_inode->i_rdev)];
  MOD_DEC_USE_COUNT;
  if (hh) {
    mem_map_dec_count(MAP_NR(hh->vme_base_p));
  }
}

/* -------- DM 11.2.99 ----------------------------------------------------- */
/*
 * this is my nopage handler, called if the proc generates
 * a page fault (i.e. the user exceeds the VMESIZE limitation
 * (s. filemap.c in /usr/src/linux/mm)
 */
static unsigned long arwvme_vma_nopage(struct vm_area_struct *vma,
                                       unsigned long address, int no_share)
{
  ArwvmeHardwareHandle *hh;
  ArwvmeUserHandle *uh;
  unsigned long offset = address - vma->vm_start;
/*
 * address points to page begin!
 */

  if (offset < 0 || address >= vma->vm_end) {
    printk("arwvme_no_page(): address 0x%lx out of range!\n",address);
    return 0L; /* SIGBUS */
  }
  hh = modnr2hh[MINOR(vma->vm_inode->i_rdev)];
  if (!hh) {
    printk("arwvme_vma_nopage(): no minor!\n");
    return 0L; /* SIGBUS */
  }
  LOCK_HH(hh);
  uh = ArwvmeGetUserHandleFromPID(hh, current->pid);
  if (!uh) {
    RELEASE_HH(hh);
    printk("arwvme_nopage(): user unknown!\n");
    return 0L; /* SIGBUS */
  }

  if (!hh->current_uh || current->pid != hh->current_uh->pid) {
    Arwvme_switch_context(hh, uh);
  }

  Arwvme_build_address(hh, uh, offset + uh->mmapd.vme_offs);

  queue_task(&hh->psn, &tq_scheduler);

  uh->mmapd.page_status = 1;

  RELEASE_HH(hh);

/* hack the kernel */
  current->mm->rss--;
 
  return (unsigned long)hh->vme_base_p;
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static int arwvme_mmap(struct inode *inode, struct file *file,
                                            struct vm_area_struct *vma )
/*
 * a call to this function will only 
 * install a new vm_operations_struct
 * in vma. This new struct has a new
 * nopage handler, which returns the 
 * VMESIZE-VMEbus page, if the user 
 * requests an access to his malloced 
 * area.
 */
{
  ArwvmeHardwareHandle *hh = modnr2hh[MINOR(inode->i_rdev)];
  ArwvmeUserHandle *uh;

  if (!hh) {
    printk("arwvme_mmap(): no such device!\n");
    return -ENODEV;
  }

  ArwvmePrint(("arwvme_mmap() ...\n"));

  uh = file->private_data;
  if (!uh) {
    return -EBADF;
  }

  if (vma->vm_offset & (PAGE_SIZE-1))
    return -ENXIO; /* need aligned offsets */

  if (vma->vm_offset != hh->vme_base_p) {
    return -ENXIO;
  }

/*
 * do nothing, nopage() will get
 * the page (hh->vme_page_p) and
 * set the VIC68 registers to the
 * actual context!
 */ 
  vma->vm_ops = &arwvme_vmops;

  vma_set_inode(vma,inode);
  inode->i_count++;

  uh->mmapd.size = vma->vm_end - vma->vm_start;
  uh->mmapd.reserved_page = vma->vm_start;
  uh->mmapd.file = file;

  ArwvmePrint(("... end of arwvme_mmap().\n"));

  arwvme_vma_open(vma);

  return 0;
}

/* low level read, write */
/* -------- DM 17.2.99 ----------------------------------------------------- */
static void ArwvmeWriteA_D8(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                            off_t f_pos, int data)
{
  void *ptr = Arwvme_build_address(hh, uh, f_pos); 
  writeb(data, ptr);
}

static void ArwvmeWriteA_D16(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                            off_t f_pos, int data)
{
  void *ptr = Arwvme_build_address(hh, uh, f_pos); 
  writew(data, ptr);
}

static void ArwvmeWriteA_D32(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                            off_t f_pos, int data)
{
  void *ptr = Arwvme_build_address(hh, uh, f_pos); 
  writel(data, ptr);
}

/* -------- DM 17.2.99 ----------------------------------------------------- */
static int ArwvmeReadA_D8(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                          off_t f_pos)
{
  void *ptr = Arwvme_build_address(hh, uh, f_pos); 
  return readb(ptr); 
}

static int ArwvmeReadA_D16(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                          off_t f_pos)
{
  void *ptr = Arwvme_build_address(hh, uh, f_pos); 
  return readw(ptr); 
}

static int ArwvmeReadA_D32(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh,
                          off_t f_pos)
{
  void *ptr = Arwvme_build_address(hh, uh, f_pos); 
  return readl(ptr); 
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static ssize_t arwvme_read(struct inode *inode, struct file *file,
                                            char *buf, int count)
{
  ArwvmeHardwareHandle *hh = modnr2hh[MINOR(inode->i_rdev)];
  unsigned int w;
  ArwvmeUserHandle *uh;
  int i=0,dtype,ret;
  int dti;

  if (!hh) {
    printk("arwvme_read(): no such device!\n");
    return -ENODEV;
  }

  if ((ret = verify_area(VERIFY_WRITE, (void *)buf, count))) {
    printk("ARWVME::VERIFY_WRITE in read(): %p for pid %d\n",
           buf,current->pid);
    return ret;
  }

  uh = (ArwvmeUserHandle *)file->private_data;
  if (!uh) {
    return -EBADF;
  }
  ArwvmePrint(("arwvme_read() for user %d ...\n", uh->pid));

/* 
 * don't get another arw user in between 
 * else we have a context problem!
 */

  if (!uh->vmed.ada.data_type) {
    printk("arwvme_read(): user %d has not chosen data type yet!\n", current->pid);
    return -EIO;
  }

  if (count > ARWVME_RW_MAX_COUNT) {
    count = ARWVME_RW_MAX_COUNT;
  }

  LOCK_HH(hh); 

/* 
 * disable an eventually used mmap 
 * whilst this schedule
 */  
  ArwvmeInvalidatePage(hh);

  hh->in_rw = TRUE;

  Arwvme_switch_context(hh, uh);

  dti = ARWVME_DATA_TYPE_INDEX(uh->vmed.ada.data_type);
  dtype = 1 << dti;

  for (i=0;i<count;i+=dtype) {
    w = ArwvmeReadVTBL[dti](hh, uh, file->f_pos); 
    if (ARWVME_GET_BUS_ERR(hh->vic_base)) {
      ArwvmePrint(("Arwvme_read(): error occured during vmebus access!\n"));
/*
 * 1. get rid of the pending int
 * if this user has disabled the irqs
 * 2. clr berr flag in VIC chip
 */
      if (!uh->intInInts) { /* else ack in int handler */
        ARWVME_READ_INT_VEC(hh->mem_base);
        ArwvmeSetSysErrInfo(hh, uh);
#if SIGBUSON        
        kill_proc(current->pid, SIGBUS, 0);
#endif
      }
      ARWVME_CLR_BESR(hh->vic_base);
      RELEASE_HH(hh);
      return i;
    }
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,4)
    __put_user(w,buf+i, dtype);
#else
    __put_user_nocheck(w,buf+i, dtype);
#endif
    if (uh->fpos_tied) 
      file->f_pos += dtype;
  }

  hh->in_rw = FALSE;

  RELEASE_HH(hh);

  ArwvmePrint(("... read %d bytes.\n", i));

  return i;
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static ssize_t arwvme_write(struct inode *inode,
                        struct file *file, const char *buf, int count) 
{
  ArwvmeHardwareHandle *hh = modnr2hh[MINOR(inode->i_rdev)];
  ArwvmeUserHandle *uh;
  int i=0,w,dtype,ret;
  int dti;

  if (!hh) {
    printk("arwvme_write(): no such device!\n");
    return -ENODEV;
  }

  if ((ret = verify_area(VERIFY_READ, (void *)buf, count))) {
    printk("ARWVME::VERIFY_READ in write(): %p for pid %d\n",buf,current->pid);
    return ret;
  }

  uh = (ArwvmeUserHandle *)file->private_data;
  if (!uh) {
    return -EBADF;
  }

  if (!uh->vmed.ada.data_type) {
    printk("arwvme_write(): user %d has not chosen data type yet!\n", current->pid);
    return -EIO;
  }

/*
 * cutting down the max. number
 * of transferred bytes to prevent
 * bus hogging.
 */
  if (count > ARWVME_RW_MAX_COUNT) {
    count = ARWVME_RW_MAX_COUNT;
  }

  ArwvmePrint(("arwvme_write() for user %d ...\n", current->pid));

/* don't get another arw user in between */
  LOCK_HH(hh);

  ArwvmeInvalidatePage(hh);

  hh->in_rw = TRUE;

  Arwvme_switch_context(hh, uh);

  dti = ARWVME_DATA_TYPE_INDEX(uh->vmed.ada.data_type);
  dtype = 1 << dti;

  for (i=0;i<count;i+=dtype) {
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,4)
    w = (unsigned int)__get_user(buf+i, dtype);
#else
    __get_user_nocheck(w,buf+i, dtype);
#endif
    ArwvmeWriteVTBL[dti](hh, uh, file->f_pos, w); 
    if (ARWVME_GET_BUS_ERR(hh->vic_base)) {
      ArwvmePrint(("Arwvme_write(): error occured during vmebus access!\n"));
      if (!uh->intInInts) { /* else ack in int handler */
        ARWVME_READ_INT_VEC(hh->mem_base);
        ArwvmeSetSysErrInfo(hh, uh);
#if SIGBUSON        
        kill_proc(current->pid, SIGBUS, 0);
#endif
      }
      ARWVME_CLR_BESR(hh->vic_base);
      RELEASE_HH(hh);
      return i;
    }
    if (uh->fpos_tied) file->f_pos += dtype;
  }

  hh->in_rw = FALSE;

  RELEASE_HH(hh);

  ArwvmePrint(("... written %d bytes.\n", i));

  return i;
}

static int ArwvmePutKernelValue(int cmd, void *to, void *from)
{
  int size = _IOC_SIZE(cmd);
  int ret;
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,4)
         if ((ret = verify_area(VERIFY_WRITE, (void *)to, size))) {
           printk("ARWVME::VERIFY_WRITE: %p in cmd (%d,0x%x) for pid %d\n",
           to, _IOC_DIR(cmd), _IOC_NR(cmd), current->pid);
           return ret;
         }
         memcpy_tofs((void*)to, from, size);
         return ret;
#else
         if (copy_to_user(to, from, size)) {
           printk("ARWVME::VERIFY_WRITE: %p for pid %d\n",to,current->pid);
           return -EINVAL;
         }
#endif
}

static int ArwvmeGetUserValue(int cmd, void *to, void *from)
{
  int size = _IOC_SIZE(cmd);
  int ret;
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,4)
  if ((ret = verify_area(VERIFY_READ, (void *)from, size))) {
    printk("ARWVME::VERIFY_READ %p in cmd (%d,0x%x) for pid %d\n",
           from, _IOC_DIR(cmd), _IOC_NR(cmd), current->pid);
    return ret;
  }
  memcpy_fromfs(to, from, size);
  return ret;
#else
  if (copy_from_user(to, from, size)) {
    printk("ARWVME::VERIFY_READ: %p in cmd %d for pid %d\n",from,current->pid);
    return -EFAULT;
  }
#endif
}

/* -------- DM 9.2.99 ------------------------------------------------------ */
static int arwvme_ioctl(struct inode* inode, struct file* file,
                        unsigned int cmd, unsigned long arg)
{
  ArwvmeADA ada;
  ArwvmeArbiter arbiter;
  ArwvmeIntAck user_int_ack;
  int vec, ret, vme_win, vme_to;
  ArwvmeHardwareHandle *hh = modnr2hh[MINOR(inode->i_rdev)];
  ArwvmeUserHandle *uh;
  Node *node;
  int am;
  unsigned long flags;

  if (!hh) {
    printk("arwvme_ioctl(): no such device!\n");
    return -ENODEV;
  }

  uh = (ArwvmeUserHandle *)file->private_data;
  hh->current_uh = uh;

  if (!uh) {
    printk("arwvme_ioctl(): no user handle for this pid %d!\n",current->pid);
    return -EBADF;
  }

  ArwvmePrint(("arwvme_ioctl() for user %d with cmd 0x%x ...\n", uh->pid, cmd));

  LOCK_HH(hh);

  switch (cmd) {
    case ARWVME_INITIALIZE        :
/* 
 * reserved for superuser and the group members of
 * this device.
 */
         if (!suser() && !in_group_p(inode->i_gid)) {
           RELEASE_HH(hh);
           return -EPERM;
         }
         Arwvme_init(hh);
         ArwvmeUpdateUserHandleFromHardware(hh,uh);
         RELEASE_HH(hh);
         return 0;
    case ARWVME_SYSRESET          :
/* 
 * reserved for superuser and the group members of
 * this device.
 */
         if (!suser() && !in_group_p(inode->i_gid)) {
           RELEASE_HH(hh);
           return -EPERM;
         }
         Arwvme_reset(hh);
         ArwvmeUpdateUserHandleFromHardware(hh,uh);
         RELEASE_HH(hh);
         return 0;
    case ARWVME_CHECK_CONNECTION  :
         RELEASE_HH(hh);
         ret = Arwvme_check_connection(hh);
         put_fs_long(ret, (unsigned int *)arg);
         return 0;
    case ARWVME_SET_AM            :
         if ((ret = ArwvmeGetUserValue(cmd, &ada, (void *)arg))) {
           RELEASE_HH(hh); 
           return ret;
         }
         if ((am = Arwvme_set_am(hh, uh, &ada)) < 0) {
           RELEASE_HH(hh);
           printk("ARWVME::ioctl(): no addr %x, data %x, acc %x\n",
                  ada.addr_type, ada.data_type, ada.acc_type);
           return -EINVAL;
         }
         ArwvmeUpdateUserHandleFromAM(uh, am);
         RELEASE_HH(hh);
         return 0;
    case ARWVME_ENABLE_INTERRUPT  :
         /* dummy: interrupts always enabled */
         RELEASE_HH(hh);
         return 0;
    case ARWVME_DISABLE_INTERRUPT :
         /* dummy: interrupt freed at cleanup */ 
         RELEASE_HH(hh);
         return 0;
    case ARWVME_POST_INTERRUPT     :
         RELEASE_HH(hh);
         if (!uh->int_ack) {
           uh->int_ack = ArwvmeAllocIntAckVector();
           if (!uh->int_ack) {
             return -ENOMEM;
           }
         }
         if ((ret = ArwvmeGetUserValue(cmd, &user_int_ack, (void *)arg))) return ret;
         vec = user_int_ack.int_vec;
         if (vec <= ARWVME_MAX_VECTOR || vec >= 0x100) { 
           printk("ioctl()::ARWVME_POST_INTERRUPT: bad int vec %x\n!",vec);
           return -EINVAL;
         }
         ArwvmePrint(("... user %d interrested in vector %d.\n", uh->pid, vec));
         memcpy(uh->int_ack+vec,&user_int_ack,sizeof(ArwvmeIntAck));
         uh->intInInts = TRUE;
         return 0;
    case ARWVME_REMOVE_INTERRUPT  :
         RELEASE_HH(hh);
         if ((ret = ArwvmeGetUserValue(cmd, &vec, (void *)arg))) return ret;
         if (vec <= ARWVME_MAX_VECTOR || vec >= 0x100) 
           printk("ioctl()::ARWVME_POST_INTERRUPT: bad int vec %x\n!",vec);
           return -EINVAL;
         uh->int_ack[vec].int_vec = -1;
         ArwvmePrint(("... user %d not interrested in vector %d.\n", uh->pid,
                      vec));
         return 0;   
    case ARWVME_GET_INTERRUPT:
         RELEASE_HH(hh);
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,4)
         if ((ret = verify_area(VERIFY_WRITE, (void *)arg, sizeof(ArwvmeIntAck)))) {
           printk("ARWVME::VERIFY_WRITE: %ld for pid %d\n",arg,current->pid);
           return ret;
         }
#endif

/*
 * do NOT exchange data between user and kernel
 * space whilst in cli()/sti(), 'cos the user
 * page may be swapped out and so the process
 * must deschedule for demanding the page from
 * fs. 
 */
         save_flags(flags);
         cli();

         if (!uh->pendingInts) {
           restore_flags(flags);
           return -EINVAL;
         }
         node = getFirstNode(uh->pendingInts);
         if (!node) {
           restore_flags(flags);
           return -EINVAL;
         } 
         memcpy(&user_int_ack, node->data, sizeof(ArwvmeIntAck));
         delFirstNodeAndData(uh->pendingInts);
         restore_flags(flags);

         ArwvmePrint(("... retrieved interrupt with vec 0x%x and tl %Ld.\n", 
                      user_int_ack.int_vec, user_int_ack.int_tl));
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,4)
         memcpy_tofs((void *)arg, &user_int_ack, sizeof(ArwvmeIntAck));
#else
         if (copy_to_user((void *)arg, &user_int_ack, sizeof(ArwvmeIntAck))) {
           printk("ARWVME::VERIFY_WRITE: %ld for pid %d\n",arg,current->pid);
           return -EINVAL;
         } 
#endif
         return 0;
    case ARWVME_ATTACH_SIGNAL     :
         RELEASE_HH(hh);
         return ArwvmeGetUserValue(cmd, &(uh->s_pid), (void *)arg);
    case ARWVME_SET_VME_WINDOW    :
         if ((ret = ArwvmeGetUserValue(cmd, &vme_win, (void *)arg))) {
           RELEASE_HH(hh);
           return ret;
         }
         Arwvme_build_address(hh, uh, vme_win);
         RELEASE_HH(hh);
         uh->mmapd.vme_offs = uh->vmed.vme_addr;
         ArwvmePrint(("... set VME Window to 0x%x.\n", uh->mmapd.vme_offs));
         return 0;
    case ARWVME_GET_DRIVER_INFO   :
         RELEASE_HH(hh);
         if (!hh->driver_info) {
           return -EINVAL;
         }
         return ArwvmePutKernelValue(cmd, (void *)arg, hh->driver_info);
    case ARWVME_ERROR_CHECK       :
         if ((ret = ArwvmePutKernelValue(cmd, (void *)arg, &uh->sys_err))) {
           RELEASE_HH(hh);
           return ret;
         }
         ArwvmeResetSysErrInfo(uh);
         RELEASE_HH(hh);
         return 0;
    case ARWVME_TIE_FPOS          :
         RELEASE_HH(hh);
         ArwvmePrint(("... ARWVME_TIE_FPOS.\n"));
         uh->fpos_tied = TRUE;
         return 0; 
    case ARWVME_UNTIE_FPOS        :
         RELEASE_HH(hh);
         ArwvmePrint(("... ARWVME_UNTIE_FPOS.\n"));
         uh->fpos_tied = FALSE;
         return 0; 
    case ARWVME_GET_VME_SIZE      :
         RELEASE_HH(hh);
         return ArwvmePutKernelValue(cmd, (void *)arg, &hh->vme_range);
    case ARWVME_GET_VME_BASE      :
         RELEASE_HH(hh);
         return ArwvmePutKernelValue(cmd, (void *)arg, &hh->vme_base_p);
    case ARWVME_CTRL_ARBITER      :
         if ((ret = ArwvmeGetUserValue(cmd, &arbiter, (void *)arg))) {
           RELEASE_HH(hh);
           return ret;
         } 
         if (arbiter.arb_mode < 0 || arbiter.arb_mode > 1) {
           RELEASE_HH(hh);
           return -EINVAL;
         }
         if (arbiter.br_lev < 0 || arbiter.br_lev > 3) {
           RELEASE_HH(hh);
           return -EINVAL;
         }
         if (arbiter.fairness < 0 || arbiter.fairness > 0xF) {
           RELEASE_HH(hh);
           return -EINVAL;
         }
         uh->vmed.arbiter = (arbiter.arb_mode << 7) + 
                            (arbiter.br_lev   << 5) +
                            arbiter.fairness; 
         RELEASE_HH(hh);
         return 0;
    case ARWVME_SET_VMEBUS_TO     :
         if ((ret = ArwvmeGetUserValue(cmd, &vme_to, (void *)arg))) {
           RELEASE_HH(hh);
           return ret;
         } 
         if (arg < 0 || arg > 0x7) {
           RELEASE_HH(hh);
           return -EINVAL;
         }
/* set only 7:5 */
         B_COPY_BITS(&uh->vmed.vmebus_to, (unsigned char)vme_to, 5, 7); 
         RELEASE_HH(hh);
         return 0;
    default                       :
         RELEASE_HH(hh);
         printk("ARWVME::ioctl(): unknown command!\n");
         return -EINVAL;
  }
  return -EINVAL;
}

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,4)
static struct file_operations arwvme_fops =      /* new VFS */
{
        arwvme_lseek,    /* Seek */
        arwvme_read,     /* Read */
        arwvme_write,    /* Write */
        NULL,            /* Readdir */
        arwvme_poll,     /* Poll */
        arwvme_ioctl,    /* Ioctl */
        arwvme_mmap,     /* MMap */
        arwvme_open,     /* Open */
        NULL,            /* Flush */
        arwvme_release,  /* Release */
        NULL,            /* Fsync */ 
        NULL,            /* Fasync */
        NULL,            /* Check_media_change */ 
        NULL,            /* Revalidate */ 
        NULL             /* Lock */
};
#else
static struct file_operations arwvme_fops =      /* old VFS */
{
        arwvme_lseek,    /* Seek */
        arwvme_read,     /* Read */
        arwvme_write,    /* Write */
        NULL,            /* Readdir */
        arwvme_select,   /* Select */
        arwvme_ioctl,    /* Ioctl */
        arwvme_mmap,     /* MMap */
        arwvme_open,     /* Open */
        arwvme_release,  /* Release */
        NULL,            /* Fsync */
        NULL,            /* Fasync */
        NULL,            /* Check_media_change */ 
        NULL             /* Revalidate */ 
};
#endif

/* -------- DM 9.2.99 ------------------------------------------------------ */
static int Arwvme_check_connection(ArwvmeHardwareHandle *hh)
{
  int m;
  unsigned int lcr_base = ARWVME_LCR_BASE(hh);

  ArwvmePrint(("Arwvme_check_connection() for index %d ...\n", hh->index));

  ARWVME_EN_VMEMM(lcr_base);

  if (!ARWVME_VMEMM_OK(lcr_base)) {
    ArwvmePrint(("Arwvme_check_connection(): can't access VMEMM %d!\n", hh->index));
    return 0;
  }

  ARWVME_CLR_INT2(lcr_base);
  ARWVME_EN_INT(lcr_base);

  hh->module_number = ARWVME_GET_MODULE_NUMBER(hh->mem_base);
  if (ARWVME_TST_PCIADA_INT(lcr_base)) {
    ARWVME_CLR_INT2(lcr_base);
    printk("Arwvme_check_connection(): PCIADA timeout interrupt occured -> no access to VMEMM!\n");
    return 0;
  }

  ArwvmePrint(("Arwvme_check_connection(): found module number %d for index %d!\n", hh->module_number, hh->index));

  if ((m = ARWVME_GET_MODULE_TYPE_ID(hh->mem_base)) != 0x1) {
    ArwvmePrint(("Arwvme_check_connection(): wrong module type id %d (should be %d)!\n",m,ARWVME_MODTYPE));
    return 0;
  }
  ArwvmePrint(("... connection ok for index %d with module number %d.\n", 
               hh->index, hh->module_number)); 
  if (modnr2hh[hh->module_number]) {
    if (hh->module_number != modnr2hh[hh->module_number]->module_number) {
       ArwvmePrint(("Arwvme_check_connection(): pci index %d and %d have the same module number ...\n"));
       ArwvmePrint(("... this module number will be removed!!\n"));
       return 0;
    }
  } 
  modnr2hh[hh->module_number] = hh;
  return 1;
}

/* -------- DM 4.3.99 ------------------------------------------------------ */
static void ArwvmeFreeUserHandle(ArwvmeUserHandle *uh)
{
  if (uh->pendingInts) deleteList(uh->pendingInts,NULL);
  if (uh->int_ack) vfree(uh->int_ack);
  FREE(uh);
}

/* -------- DM 4.3.99 ------------------------------------------------------ */
static void ArwvmeFreeHardwareHandle(ArwvmeHardwareHandle *hh)
{
  if (hh->lcr_base) iounmap((void *)hh->lcr_base);
  if (hh->driver_info) FREE(hh->driver_info); 
  if (hh->mm_user) deleteList(hh->mm_user,(void (*)(void *))ArwvmeFreeUserHandle);
  release_region(hh->io_base, ARWVME_IO_LENGTH);
  FREE(hh);
}

/* -------- DM 12.5.99 ----------------------------------------------------- */
static void ArwvmePrintPageSetup(unsigned long addr)
{
  mem_map_t *p = mem_map + MAP_NR(addr);
  printk("mem_map[%ld]:\n", MAP_NR(addr));
  printk("next:  %p\n", p->next); 
  printk("prev:  %p\n", p->prev); 
  printk("inode: %p\n",p->inode);
  printk("offset: 0x%lx\n",p->offset);
  printk("count: %d\n",p->count);
  printk("flags: 0x%x\n",p->flags); 
  printk("wait: %p\n",p->wait);
  printk("buffers: %p\n",p->buffers);
}

/* -------- DM 12.5.99 ----------------------------------------------------- */
static void ArwvmeSetupPage(unsigned long addr)
{
  mem_map_t *p = mem_map + MAP_NR(addr);
  memset(p, 0, sizeof(mem_map_t));
  p->flags = PG_DMA|PG_reserved;
  p->map_nr = p - mem_map; 
}

/* -------- DM 9.2.99 ------------------------------------------------------ */
static List *ArwvmeSetupHardwareHandle(List *pch)
/* builds a list of hardware informations */
/* from the pci_config_header list        */
{
  List *pbas;
  Node *node;
  ArwvmeHardwareHandle *p;
  PCIConfigHeader *pci_conf;
  int o;

  if (getNumOfNodesInList(pch) < 1) {
    printk("ArwvmeSetupHardwareHandle(): no hardware!\n");
    return NULL;
  }
  pbas = newList();
  if (!pbas) {
    printk("ArwvmeSetupHardwareHandle(): no memory!\n");
    return NULL;
  }

  node = getFirstNode(pch);
  while (node) {
    p = ANEW(ArwvmeHardwareHandle); 
    if (!p) {
      printk("ArwvmeSetupHardwareHandle(): no memory!\n");
      return pbas;
    }
    pci_conf = (PCIConfigHeader *)node->data;
/* memory mapped access to PLX Configuration Registers 0x10 */
    p->module_number = -1;

/*
 * There is a nice bug in the PLX-chip (Rev. 1):
 *  if bit 7 of i/o- or mmapped PCR is set
 *  you can't read the registers, so there is
 *  a second set of PCR allocated
 *  from EPROMed startup sequence.
 */
/* in some revisions of PLX9050 is a bug, disabling access to LCR */ 
    p->lcr_base = ARWVME_PCIBAR0(pci_conf->base_address[0]);
    if (p->lcr_base & (1<<7)) { /* buf fixup for mmap'ed LCR */
      printk("ArwvmeSetupHardwareHandle(): performing fix for mmap'ed LCR for index %d\n",
             pci_conf->index);
      pcibios_read_config_dword(pci_conf->busnr, pci_conf->devnr,PCI_BASE_ADDRESS_4,pci_conf->base_address+4);
      pcibios_write_config_dword(pci_conf->busnr, pci_conf->devnr,PCI_BASE_ADDRESS_0,pci_conf->base_address[4]);
      p->lcr_base =  ARWVME_PCIBAR1(pci_conf->base_address[0]);
    }
/* i/o-mapped access to PLX Configuration Registers 0x14 */
    p->io_base =  ARWVME_PCIBAR1(pci_conf->base_address[1]);
    if (p->io_base & (1<<7)) {
      printk("ArwvmeSetupHardwareHandle(): performing fix mmap'ed LCR for index %d\n",
             pci_conf->index);
      pcibios_read_config_dword(pci_conf->busnr, pci_conf->devnr,PCI_BASE_ADDRESS_5,pci_conf->base_address+5);
      pcibios_write_config_dword(pci_conf->busnr, pci_conf->devnr,PCI_BASE_ADDRESS_1,pci_conf->base_address[5]);
      p->io_base =  ARWVME_PCIBAR1(pci_conf->base_address[1]);
    }
/* Local Address Space 0 0x18 */
    p->mem_base = ARWVME_PCIBAR2(pci_conf->base_address[2]); 
    ArwvmePrint(("ArwvmeSetupHardwareHandle(): mem_base = 0x%x\n",p->mem_base));

    if (check_region(p->lcr_base, ARWVME_IO_LENGTH)) {
      printk("ArwvmeSetupHardwareHandle(): mmapped LCR region 0x%X does not check for PCI index %d!\n", 
             p->io_base, p->pci_conf->index);
      FREE(p);
      node = node->succ;
      continue;
    } 
/* allow kernel access to mem > 1MB */
    p->lcr_base = (unsigned int)ioremap(p->lcr_base, ARWVME_IO_LENGTH);
    if (!p->lcr_base) {
      printk("ArwvmeSetupHardwareHandle(): can't remap lcr_base into kernel!\n");
      FREE(p);
      node = node->succ;
      continue;
    }

    p->mem_range = ARWVME_LAS0RR(readl_reg(PLX9050_LAS0RR+ARWVME_LCR_BASE(p)));
    if (check_region(p->mem_base, p->mem_range)) {
      printk("ArwvmeSetupHardwareHandle(): VMEMM-region 0x%lx, 0x%x does not check for PCI index %d!\n", 
             p->mem_base, p->mem_range,
             p->pci_conf->index);
      FREE(p);
      node = node->succ;
      continue;
    } 

    p->pci_conf = pci_conf;
    if (check_region(p->io_base, ARWVME_IO_LENGTH)) {
      printk("ArwvmeSetupHardwareHandle(): I/O-region 0x%X does not check for PCI index %d!\n", 
             p->io_base, p->pci_conf->index);
      FREE(p);
      node = node->succ;
      continue;
    } 

    p->initiated = FALSE;

    p->magic = ARWVME_MAGIC;

    p->mem_base = (unsigned long)bus_to_virt(p->mem_base);
    p->vic_base = ARWVME_VIC_BASE(p->mem_base);
    p->vec_base = ARWVME_VEC_BASE(p->mem_base);
    p->vme_base = ARWVME_VME_BASE(p->mem_base);
    p->vme_base_p = (unsigned long)virt_to_phys((void *)p->vme_base);
    if (p->vme_base_p & (PAGE_SIZE-1)) {
      printk("ArwvmeSetupHardwareHandle(): VME window not on page boundary!\n");
      FREE(p);
      node = node->succ;
      continue;
    }
/* there will be a camac controller with 32K */
    p->vme_end =  p->mem_base + p->mem_range;
    p->vme_range = p->vme_end - p->vme_base;

    p->index = p->pci_conf->index; 

    if (!Arwvme_check_connection(p)) {
      printk("ArwvmeSetupHardwareHandle(): no connection to VMEMM of index %d\n",
             p->index);
      FREE(p);
      node = node->succ;
      continue;
    }

    p->word_path = (ARWVME_TEST_J401(p->mem_base) ? /* bus-width check for VMEbus */
                     ARWVME_VMEMM3 : ARWVME_VMEMM6); 
    if (!ARWVME_TEST_J402(p->mem_base)) { /* system-controller check */
      printk("ArwvmeSetupHardwareHandle(): this board is NOT jumpered for system-controller!\n");
      FREE(p);
      return NULL;
    }
    p->mm_user = newList();
    p->current_uh = NULL;
    p->driver_info = kmalloc(ARWVME_DRIVER_INFO_SIZE, GFP_KERNEL);
    if (p->driver_info) {
      o = sprintf(p->driver_info, "%s%d driver initialized with\n",
                  ARWVME_NAME, p->index);
      o += sprintf(p->driver_info+o, "VIC_BASE 0x%lx\n", p->vic_base);
      o += sprintf(p->driver_info+o, "VME_BASE 0x%lx\n", p->vme_base);
      o += sprintf(p->driver_info+o, "VME_END 0x%lx\n", p->vme_end);
      o += sprintf(p->driver_info+o, "MEM_RANGE 0x%x\n", p->mem_range);
      o += sprintf(p->driver_info+o, "VME_RANGE 0x%x\n", p->vme_range);
      o += sprintf(p->driver_info+o, "Board %s\n",
                 (p->word_path == ARWVME_VMEMM3 ? 
                   "3HE Board (only D16)" : "6HE Board"));
      o += sprintf(p->driver_info+o, "%s\n",
                  "$Revision: 1.37 $ $Date: 1999/07/13 08:44:45 $");
      ArwvmePrint(("ArwvmeSetupHardwareHandle(): \n%s\n", p->driver_info));
    }
    p->in_rw = FALSE;
    p->int_vec = 0;
    p->irq = 0;
    p->wait_queue = NULL; 
    p->sem = MUTEX;

/*
 * InvalidatePage() stuff
 */
    p->psn.next = NULL;
    p->psn.sync = 0;
    p->psn.routine = (void (*)(void *))ArwvmeInvalidatePage;
    p->psn.data = p;

    node = node->succ;
    addTail(pbas,p);
  }
  return pbas;
}

/* -------- DM 17.2.99 ----------------------------------------------------- */
void arwvme_setup(char *str, int *ints)
{
/* we need no config data here */
  return;
}

/* -------- DM 9.2.99 ------------------------------------------------------ */
void arwvme_init(void)
{
  init_module();
}

/* -------- DM 9.2.99 ------------------------------------------------------ */
void cleanup_module(void)
{
  ArwvmeHardwareHandle *hh; 
  Node *hhn;

  if (MOD_IN_USE) {
    printk("ARWVME::cleanup_module(): device is busy!\n");
    return;
  }

  printk("ARWVME::cleanup_module(): try to remove module ...\n");
  hhn = getFirstNode(hardwareHandle);
  while (hhn) {
    hh = (ArwvmeHardwareHandle *)(hhn->data);
    if (hh->initiated) Arwvme_deinit(hh);
    hhn = hhn->succ;  
  }

  deleteList(pci_device_header,NULL);
  deleteList(hardwareHandle,(void (*)(void *))ArwvmeFreeHardwareHandle);

  if (unregister_chrdev(ARWVME_MAJOR, ARWVME_NAME)) {
    printk("ARWVME::cleanup_module(): failed!\n");
  } else {
    printk("ARWVME::cleanup_module(): module removed!\n");
  }
}

/* -------- DM 17.2.99 ----------------------------------------------------- */
static int ArwvmeRequestIOPorts(ArwvmeHardwareHandle *hh) 
{
  if (check_region(hh->io_base, ARWVME_IO_LENGTH)) {
    printk("unable to access i/o address 0x%04X\n", hh->io_base);    
    return 0;
  }

  printk("%s: Ports 0x%04X - 0x%04X have been reserved.\n",
          ARWVME_NAME, hh->io_base, hh->io_base + ARWVME_IO_LENGTH - 1);
  request_region(hh->io_base, ARWVME_IO_LENGTH, ARWVME_NAME);

  return 1;
}
  
/* -------- DM 9.2.99 ------------------------------------------------------ */
int init_module(void)
{
  PCIConfigHeader *ch;
  Node *pci_ban;
  ArwvmeHardwareHandle *hh; 
  int i = 0;

#if LINUX_VERSION_CODE < LINUX_NO_SHIRQ
  for (i=0;i<NUMBER(irq2hh);i++) irq2hh[i] = NULL;
#endif
  for (i=0;i<NUMBER(modnr2hh);i++) modnr2hh[i] = NULL;

#ifdef LCR_MMAPPED
  ArwvmePrint(("init_module(): module compiled for mmapped access to LCR!\n"));
#else
  ArwvmePrint(("init_module(): module compiled for io access to LCR!\n"));
#endif

  if (pci_device_header) resetList(pci_device_header,NULL);
  else pci_device_header = newList();

  if (hardwareHandle) resetList(hardwareHandle,(void (*)(void *))ArwvmeFreeHardwareHandle);
  else hardwareHandle = newList();

  i = 0;
  while ((ch = GetPCIConfigHeader(ARWVME_VENDOR_ID, ARWVME_DEVICE_ID, i))) {
    if (ch->class_code != ARWVME_CLASS_CODE) {
      printk("ARVME::init_module(): found 0x%x, 0x%x , but wrong class code %x!\n",ARWVME_VENDOR_ID, ARWVME_DEVICE_ID, ch->class_code);
      i++;
      continue;
    }
    if (ch->subsystem_id != ARWVME_SUBSYSTEM_ID) {
      printk("ARVME::init_module(): found 0x%x, 0x%x , but wrong subsystem id %x!\n",ARWVME_VENDOR_ID, ARWVME_DEVICE_ID, ch->subsystem_id);
      i++;
      continue;
    }
    if (ch->subsystem_vendor_id != ARWVME_SUBSYSTEM_VENDOR_ID) {
      printk("ARVME::init_module(): found 0x%x, 0x%x , but wrong subsystem vendor Id %x!\n",ARWVME_VENDOR_ID, ARWVME_DEVICE_ID, ch->subsystem_vendor_id);
      i++;
      continue;
    }
    ArwvmePrint(("ARWVME::init_module() found hardware in index %d ...\n", i));
    addTail(pci_device_header,ch);
    i++;
  }
  if (!i) {
    printk("ARVME::init_module(): no PCIADA with VendorID 0x%x and DeviceId 0x%x found!\n", ARWVME_VENDOR_ID, ARWVME_DEVICE_ID);
    return -ENODEV;
  }
  ArwvmePrint(("... found %d PCIADAs!\n",i));

  printk("pci_device_header->nodes: %d\n",pci_device_header->nodes);
  hardwareHandle = ArwvmeSetupHardwareHandle(pci_device_header);
  if (!hardwareHandle) {
    printk("ARWVME::init_module(): can't access hardware config data!\n");
    return -EIO;
  }

  if (register_chrdev(ARWVME_MAJOR, ARWVME_NAME, &arwvme_fops)) {
    printk("ARWVME::init_module(): unable to get major %i for %s device\n",
           ARWVME_MAJOR, ARWVME_NAME);
    return -EIO;
  }
  ArwvmePrint(("... registered %d %s ...\n", ARWVME_MAJOR, ARWVME_NAME));

  pci_ban = getFirstNode(hardwareHandle);

/* now initialize VMEMM */
  while (pci_ban) {
    hh = (ArwvmeHardwareHandle *)(pci_ban->data); 
    if (!ArwvmeRequestIOPorts(hh)) {
      printk("ARWVME::init_module(): failed to get access to the i/o-ports!\n");
      ArwvmeFreeHardwareHandle(hh);
      if (!pci_ban->succ) {
        delNodeInList(hardwareHandle,pci_ban);
        break;
      } else {
        pci_ban = pci_ban->succ; 
        delNodeInList(hardwareHandle,pci_ban->pred);
      }
      continue;
    }
/* init all VIC68 chips found */
    Arwvme_reset(hh);
    pci_ban = pci_ban->succ; 
  }
  if (getNumOfNodesInList(hardwareHandle) < 1) {
    printk("... no VMEMM reached, removing module!\n");
    unregister_chrdev(ARWVME_MAJOR, ARWVME_NAME);
    FREE(hardwareHandle);
    deleteList(pci_device_header,NULL);
    return -EIO;
  }
  ArwvmePrint(("... end of ARWVME::init_module().\n"));
  return 0;
}

/* -------- DM 8.2.99 ------------------------------------------------------ */
static void Arwvme_init(ArwvmeHardwareHandle *hh)
{
  unsigned int lcr_base = ARWVME_LCR_BASE(hh);

  if (hh->initiated) return;

  ArwvmePrint(("Arwvme_init() for index %d ...\n", hh->index));
/*
 * Initialization of hardware
 * cut from A. Rausch: PCI-Bus to ...
 */
/* disable the 2 interrupt sources */
  writew_reg(0x0,ARWVME_INTCSR(lcr_base));
/* enable VMEMM and clr int 2 */
  ARWVME_CLR_INT2(lcr_base);

/*
 * Enables/Disables local signaling of ...
 */
  /* ... VMEbus irq ack cycles (na) ... */
  writeb(0xf8+1, hh->vic_base+VIC68_VIICR);

  /* ... VMEbus irqs (a) ... */
  writeb(0x78+1, hh->vic_base+VIC68_VICR(1));
  writeb(0x78+2, hh->vic_base+VIC68_VICR(2));
  writeb(0x78+3, hh->vic_base+VIC68_VICR(3));
  writeb(0x78+4, hh->vic_base+VIC68_VICR(4));
  writeb(0x78+5, hh->vic_base+VIC68_VICR(5));
  writeb(0x78+6, hh->vic_base+VIC68_VICR(6));
  writeb(0x78+7, hh->vic_base+VIC68_VICR(7));

  /* ... DMA interrupt on completion (na) ... */
  writeb(0xf8+0, hh->vic_base+VIC68_DMASICR);

  /* ... LIRQs, triggering conditions (na) ... */
  writeb(0xf8+1, hh->vic_base+VIC68_LICR(1));
  writeb(0xf8+2, hh->vic_base+VIC68_LICR(2));
  writeb(0xf8+3, hh->vic_base+VIC68_LICR(3));
  writeb(0xf8+4, hh->vic_base+VIC68_LICR(4));
  writeb(0xf8+5, hh->vic_base+VIC68_LICR(5));
  writeb(0x38+6, hh->vic_base+VIC68_LICR(6));
  writeb(0x38+7, hh->vic_base+VIC68_LICR(7));

  /* ... Global Switch activation (na) ... */
  writeb(0xf8+2, hh->vic_base+VIC68_ICGSICR);

  /* ... Module Switch activation (na) ... */
  writeb(0xf8+3, hh->vic_base+VIC68_ICMSICR);

  /* ... error group interrupts ... */
  /* SYSFAIL and ACFAIL active */
  writeb(0x68+6, hh->vic_base+VIC68_EGICR);

/*
 * ... and associated IPL. 
 */
  
/* 
 * Local ID vector for ...
 */

  /* ... Global Switches */
  /* Vec. 8 .. 11 */
  writeb(0x08, hh->vic_base+VIC68_ICGSVBR);

  /* ... Module Switches */
  /* Vec. 12 .. 15 */
  writeb(0x0c, hh->vic_base+VIC68_ICMSVBR);

  /* ... LIRQs */
  writeb(0x0, hh->vic_base+VIC68_LIVBR);

  /* ... Error Group interrupts */
  /* ACFAIL has Vec. 16 */
  writeb(0x10, hh->vic_base+VIC68_EGIVBR);

  /* Global and Module Switches register */
  /* (na) */
  writeb(0x0, hh->vic_base+VIC68_ICSR);

  /* General Purpose Interprocessor register */
  writeb(0x0, hh->vic_base+VIC68_ICR(0));
  writeb(0x0, hh->vic_base+VIC68_ICR(1));
  writeb(0x0, hh->vic_base+VIC68_ICR(2));
  writeb(0x0, hh->vic_base+VIC68_ICR(3));
  writeb(0x0, hh->vic_base+VIC68_ICR(4));

  /* VMEbus irqs activation and status */
  writeb(0xfe, hh->vic_base+VIC68_VIRSR);

  /* VMEbus irqs ID vectors */
  /* all 7 levels on Vec. 15 */
  writeb(0x0f, hh->vic_base+VIC68_VIVBR(1));
  writeb(0x0f, hh->vic_base+VIC68_VIVBR(2));
  writeb(0x0f, hh->vic_base+VIC68_VIVBR(3));
  writeb(0x0f, hh->vic_base+VIC68_VIVBR(4));
  writeb(0x0f, hh->vic_base+VIC68_VIVBR(5));
  writeb(0x0f, hh->vic_base+VIC68_VIVBR(6));
  writeb(0x0f, hh->vic_base+VIC68_VIVBR(7));

  /* Transfer TO conditions and values */
  writeb(0x3c, hh->vic_base+VIC68_TTR);         /* 16 us */

  /* Arbitration and Bus control */
  writeb(0x60, hh->vic_base+VIC68_ARCR);

  /* Address Modifier */
  writeb(0x29, hh->vic_base+VIC68_AMSR);        /* A16, D16 */

  /* Burst length for block transfers */
  /* no burst mode */
  writeb(0x0, hh->vic_base+VIC68_RCR);

  /* Reset control and semaphores for ICR0-4 */
  /* SYSFAIL cleared */
  writeb(0x80, hh->vic_base+VIC68_ICR(7));

  /* Enable interrupts */
  writew_reg(0x40, ARWVME_INTCSR(lcr_base));
  ArwvmePrint(("... end of Arwvme_init().\n")); 
  hh->initiated = TRUE;
}

/* -------- DM 9.2.99 ------------------------------------------------------ */
static void Arwvme_deinit(ArwvmeHardwareHandle *hh)
{
  unsigned int lcr_base = ARWVME_LCR_BASE(hh);
  ArwvmePrint(("Arwvme_deinit().\n"));
  /* set SYSFAIL */
  writeb(0x0, hh->vic_base+VIC68_ICR(7));

  /* disable interrupts */
  writew_reg(0x0, ARWVME_INTCSR(lcr_base));

  /* disable interface */
  writew_reg(0x4084, ARWVME_CNTRL(lcr_base));
  ARWVME_CLR_INT2(lcr_base);
  hh->initiated = FALSE;
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static int ArwvmeSearchPIDInUserList(void *d1, void *d2)
{
  ArwvmeUserHandle *uh1 = (ArwvmeUserHandle *)d1,
                   *uh2 = (ArwvmeUserHandle *)d2;
  return uh1->pid - uh2->pid;
}

/* -------- DM 2.3.99 ------------------------------------------------------ */
static ArwvmeUserHandle *ArwvmeGetUserHandleFromPID(ArwvmeHardwareHandle *hh, pid_t pid)
{
  Node *node;
  if (getNumOfNodesInList(hh->mm_user) < 1) {
    return NULL;
  }
  hh->lookup_uh.pid = pid;
  node = searchList(hh->mm_user,&hh->lookup_uh,ArwvmeSearchPIDInUserList);
  if (node) return node->data;
  return NULL;
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static ArwvmeUserHandle *ArwvmeGetUserHandleFromCurrent(ArwvmeHardwareHandle *hh)
{
  Node *node;
  if (getNumOfNodesInList(hh->mm_user) < 1) {
    return NULL;
  }
  hh->lookup_uh.pid = current->pid;
  node = searchList(hh->mm_user,&hh->lookup_uh,ArwvmeSearchPIDInUserList);
  if (node) return node->data;
  return NULL;
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static void Arwvme_reset(ArwvmeHardwareHandle *hh)
{
  unsigned char a;
  unsigned long oj;
  unsigned int lcr_base = ARWVME_LCR_BASE(hh);

  ARWVME_VIC_GLOBAL_RESET(hh->mem_base);
/*
 * the reset on the VIC68
 * lasts 200ms, so i have to
 * look, when the chip is 
 * ready to continue
 */
  oj = jiffies;
  a = 0;
  while (a != 0x29) {
    a = ARWVME_RD_ADDR_MODIFIER(hh->vic_base);
    ARWVME_CLR_INT2(lcr_base);
    ARWVME_WR_ADDR_MODIFIER(hh->vic_base, 0x29);
    ARWVME_CLR_INT2(lcr_base);
  }
  ARWVME_CLR_SYSFAIL(hh->vic_base);
#ifdef DEBUG
  printk("Resetcycle terminated after %ld ms!\n",(jiffies - oj)*10);
#endif
  hh->initiated = FALSE;

  ArwvmePrint(("Arwvme_reset() ...\n"));
  Arwvme_init(hh);
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static unsigned char Arwvme_build_am(unsigned short addr_type,
                                    unsigned short acc_type)
{
  int in, priv;
  static unsigned short ams[] = {
  /* 16 bit address */
    AM_SHORT_SUP,
    AM_SHORT_SUP,
    AM_SHORT_SUP,
    AM_SHORT_NONPRIV,
    AM_SHORT_NONPRIV,
    AM_SHORT_NONPRIV,

  /* 24 bit address */
    AM_STD_SUP_BLOCK,
    AM_STD_SUP_PROG,
    AM_STD_SUP_DATA,
    AM_STD_NONPRIV_BLOCK,
    AM_STD_NONPRIV_PROG,
    AM_STD_NONPRIV_DATA,

  /* 32 bit address */
    AM_EXT_SUP_BLOCK,
    AM_EXT_SUP_PROG,
    AM_EXT_SUP_DATA,
    AM_EXT_NONPRIV_BLOCK,
    AM_EXT_NONPRIV_PROG,
    AM_EXT_NONPRIV_DATA,
  /* 64 bit address */
    AM_LONG_SUP_BLOCK,
    AM_LONG_SUP_PROG,
    AM_LONG_SUP_DATA,
    AM_LONG_NONPRIV_BLOCK,
    AM_LONG_NONPRIV_PROG,
    AM_LONG_NONPRIV_DATA
  };

  priv = (acc_type & ARWVME_PRIV_MASK) >> 8;
  acc_type = (acc_type & ARWVME_ACC_MASK) >> 12;
  addr_type &= ARWVME_A_MASK;
  in = (addr_type-1)*6;
  in += (priv-1)*3;
  in += acc_type-1;
  if (in < 0 || in >= NUMBER(ams)) return 0;

  return ams[in];
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static int Arwvme_set_am(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh, ArwvmeADA *ada)
{
  unsigned char am;

  if (!uh) return -1;

  am = Arwvme_build_am(ada->addr_type, ada->acc_type);
  ArwvmePrint(("Arwvme_set_am(): request of am 0x%x for pid %d\n", am,current->pid));
  if (!am) return -1;
  if (am != uh->vmed.am) {
    uh->vmed.am = am;
    uh->vmed.ada.acc_type = ada->acc_type;
    uh->vmed.ada.addr_type = ada->addr_type;
    ARWVME_WR_ADDR_MODIFIER(hh->vic_base, am);
#if 0
    am = ARWVME_RD_ADDR_MODIFIER(hh->vic_base);
    ArwvmePrint(("Arwvme_set_am(): am set to 0x%x\n", am));
#endif
  }
  uh->vmed.ada.data_type = ada->data_type;
  return am;
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static void *Arwvme_build_address(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh, unsigned int addr)
/*
 * sets the VMEbus window
 * if necessary and returns
 * the physical (bus) address 
 */
{
  int offs = 0;

  if (!uh) { /* no such pid */
    return NULL;
  }

  offs = (addr % hh->vme_range);
  addr -= offs;
  if (uh->vmed.vme_addr != addr) {
    uh->vmed.vme_addr = addr;
    ARWVME_WR_VMEADDR_REG_L(hh->mem_base,addr);
    hh->current_uh = uh;
  }
  return (void *)(hh->vme_base + offs);
}

/* -------- DM 10.2.99 ----------------------------------------------------- */
static void Arwvme_switch_context(ArwvmeHardwareHandle *hh, ArwvmeUserHandle *uh)
/*
 * looking for all hardware parameters
 * the user is able to modify
 * the actual sequence is:
 * a) interrupt interest
 * b) AMSR (am)
 * c) VMEADDReg 
 * d) Arbiter
 * e) VMEbus timeout
 */
{
  unsigned int lcr_base = ARWVME_LCR_BASE(hh);
  if (getNumOfNodesInList(hh->mm_user) < 2) 
    if (hh->current_uh) return;
  /* oops, only one user on this hardware */
  /* -> nothing to swap                   */

  if (!uh->intInInts) ARWVME_DIS_INT(lcr_base);
  else ARWVME_EN_INT(lcr_base);

  if (!hh->current_uh || uh->vmed.am != hh->current_uh->vmed.am) {
    ARWVME_WR_ADDR_MODIFIER(hh->vic_base, uh->vmed.am);
  }
  if (!hh->current_uh || uh->vmed.vme_addr != hh->current_uh->vmed.vme_addr) {
    ARWVME_WR_VMEADDR_REG_L(hh->mem_base,uh->vmed.vme_addr);
  }
  if (!hh->current_uh || uh->vmed.arbiter != hh->current_uh->vmed.arbiter) {
    writeb(uh->vmed.arbiter, hh->vic_base + VIC68_ARCR);
  }
  if (!hh->current_uh || uh->vmed.vmebus_to != hh->current_uh->vmed.vmebus_to) {
    writeb(uh->vmed.vmebus_to, hh->vic_base + VIC68_TTR);
  }
  hh->current_uh = uh;
}
 
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,4)
/* -------- DM 23.6.99 ----------------------------------------------------- */
static unsigned int arwvme_poll(struct file *file, poll_table *wait)
{
  ArwvmeHardwareHandle *hh = modnr2hh[MINOR(inode->i_rdev)];
  ArwvmeUserHandle *uh;

  if (!hh) {
    printk("arwvme_ioctl(): no such device!\n");
    return POLLERR;
  }

  uh = (ArwvmeUserHandle *)file->private_data;
  if (!uh) {
    printk("arwvme_select(): no user handle for this pid %d!\n",current->pid);
    return POLLERR;
  }
  if (!uh->pendingInts) return POLLIN | POLLRDNORM; 
  /* if not interested -> don't wait */ 
  if (!emptyList(uh->pendingInts)) return POLLIN | POLLRDNORM;
  if (!ARWVME_INT_PENDING(hh->mem_base)) {
     select_wait(&hh->wait_queue, wait);
     return 0;
  }
  return POLLIN | POLLRDNORM;
}
#else
/* -------- DM 11.2.99 ----------------------------------------------------- */
static int arwvme_select(struct inode *inode, struct file *file,
                         int sel_type, select_table *wait)
{
  ArwvmeHardwareHandle *hh = modnr2hh[MINOR(inode->i_rdev)];
  ArwvmeUserHandle *uh;

  if (!hh) {
    printk("arwvme_ioctl(): no such device!\n");
    return -ENODEV;
  }

  uh = (ArwvmeUserHandle *)file->private_data;
  if (!uh) {
    printk("arwvme_select(): no user handle for this pid %d!\n",current->pid);
    return -EBADF;
  }
  if (sel_type != SEL_IN) return 1; /* hhmmm ... */
  if (!uh->pendingInts) return 1; /* if not interested -> don't wait */ 
  if (!emptyList(uh->pendingInts)) return 1;
  if (!ARWVME_INT_PENDING(hh->mem_base)) {
     select_wait(&hh->wait_queue, wait);
     return 0;
  }
  return 1;
}
#endif
