#ifndef LINUX_ARWVME_PRIV_H
#define LINUX_ARWVME_PRIV_H

#include <linux/mm.h>
#include <asm/semaphore.h>

#include "vic68.h"
#include "plx9050.h"
#include "list.h"
#include "arwpci.h"

/*
 * identifier for PCI Config header
 */
#define ARWVME_MAJOR 102
#define ARWVME_NAME  "ARW PCI-Bus to VMEbus Interface Rel 1.1"
#define ARWVME_DEVICE_ID               0x9050
#define ARWVME_VENDOR_ID               0x10B5
#define ARWVME_SUBSYSTEM_ID            0x9050
#define ARWVME_SUBSYSTEM_VENDOR_ID     0x1167
#define ARWVME_CLASS_CODE              0x801

#define ARWVME_MAGIC  ((ARWVME_VENDOR_ID<<16) | ARWVME_DEVICE_ID)

#define VALIDATE_HH(hh) (hh->magic == ARWVME_MAGIC) 

#ifdef LCR_MMAPPED
#define writel_reg(val,addr) writel(val,addr)
#define writew_reg(val,addr) writew(val,addr)
#define readl_reg(addr)      readl(addr)
#define readw_reg(addr)      readw(addr)
#define ARWVME_LCR_BASE(hh)  hh->lcr_base
#else
#define writel_reg(val,addr) outl(val,addr)
#define writew_reg(val,addr) outw(val,addr)
#define readl_reg(addr)      inl(addr)
#define readw_reg(addr)      inw(addr)
#define ARWVME_LCR_BASE(hh)  hh->io_base
#endif

/*
 * Compatibility area
 */
#define LinuxVersionCode(v, p, s) (((v)<<16)+((p)<<8)+(s))

#define LINUX_NO_SHIRQ LinuxVersionCode(1,3,70) 

#if LINUX_VERSION_CODE < LinuxVersionCode(2,0,100) 
#define ioremap vremap        
#define iounmap vfree
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE  1
#endif

#define INIT_FAIL -EIO
#define INIT_SUCCESS 0

#ifndef FAIL
#define FAIL 0
#endif

#ifndef SUCCESS
#define SUCCESS 1
#endif

#ifndef NUMBER
#define NUMBER(a) sizeof(a) / sizeof(*a)
#endif

#ifndef ANEW
#define ANEW(type) (type *)kmalloc(sizeof(type), GFP_ATOMIC)
#define FREE(ptr) kfree_s(ptr, sizeof(*ptr))
#endif

#define ARWVME_MODTYPE 1

#define ARWVME_VMEMM3 0
#define ARWVME_VMEMM6 1

#define ARWVME_IO_LENGTH        0x7F

#define ARWVME_MAX_VECTOR       0x3f

typedef struct vm_area_struct vm_area_struct;

#define ARWVME_DATA_TYPE_INDEX(dt) ((dt>>4)-1)

int B_COPY_BITS(unsigned char *target, unsigned char source, int start, int end)
{
  unsigned int mask;
  if (start > sizeof(char) || end > sizeof(char)) return 0;
  mask = (start < end) ? (end - start + 1) << start :
                         (start - end + 1) << end;
  *target &= ~mask;
  *target |= (source & mask); 
 
  return (start < end ? end - start : start - end) + 1;
}

/*
 * The VMEbus descriptor
 */
typedef struct {
  ArwvmeADA      ada;        /* address-, data- and access-type      */
  unsigned int   vme_addr;   /* actual address offset on the VMEbus  */
  unsigned char  am;         /* address modifier                     */
  unsigned char  arbiter;    /* arbiter control:                     */
                             /* mode 7:7 | br_lev 6:5 | fairness 3:0 */
  unsigned char  vmebus_to;  /* timeout value for VMEbus             */
} ArwvmeDesc;

#define ARWVME_VMED_INIT(A) memset(&(A),0,sizeof(ArwvmeDesc))

/* mmap stuff */
typedef struct {
  unsigned int vme_offs;         /* address = vme_offs + vmed.vme_addr */
  unsigned int size;             /* size of user space                 */
  unsigned long reserved_page;   /* the address of user space          */
  int page_status;               /* fully mapped ?                     */
  struct file * file;            /* the users file struct              */
                                 /* associated with this device        */
  unsigned int pg_cnt;           /* page faults cnt for this user      */
} ArwvmeMMapDesc;

#define ARWVME_MMAPD_INIT(A) memset (&(A),0,sizeof(ArwvmeMMapDesc))

/*
 * The user handle: created for a user entering the driver
 */
typedef struct {
  ArwvmeDesc     vmed;
  ArwvmeMMapDesc mmapd;
  pid_t          pid;        /* my pid                               */
  pid_t          s_pid;      /* dito, only != 0, if Signal() called  */
  int            fpos_tied;  /* flag for file pos tied to read/write */
  ArwvmeIntAck   *int_ack;   /* my posted interrupts                 */
  List           *pendingInts;   /* my received interrupts           */
  int            intInInts;      /* flag set, if some int_ack != 0   */
  ArwvmeSysErrInfo sys_err;      /* set to my (!) last bus error     */
} ArwvmeUserHandle;

/*
 * The hardware handle: created for each PCIADA/VMEMM combo
 */
typedef struct {
  unsigned int magic;        /* to identify my devices in case   */
                             /* of shared interrupts             */
  int index;                 /* attention: PCI index !           */
  int module_number;         /* the number jumpered on the VMEMM */
  unsigned int lcr_base;     /* PLX 9050 regs in PCI-space       */
  unsigned int io_base;      /* PLX 9050 regs in IO-space        */
  unsigned long mem_base,    /* MM, u.a. VIC68a                  */
               vic_base,     /* VIC68a                           */
               vec_base,     /* Interrupt vectors                */
               vme_base,     /* VME-Window (4K,32K later)        */
               vme_base_p,   /* VME-Window (phys. addr for mmap) */
               vme_end;      /* end of VME-Window                */
  unsigned int mem_range;    /* 4K or 32K + 4K                   */
  unsigned int vme_range;    /* 4K or 32K                        */
  int word_path;             /* HE3 or HE6 ?                     */
  int initiated;             /* flag for init state of VIC86     */
  PCIConfigHeader *pci_conf; /* s. arwpci.h                      */
  char *driver_info;         /* description about this driver    */
  struct _list *mm_user;     /* list of UserHandles using this module */
  ArwvmeUserHandle *current_uh,  /* last user in driver          */
                   lookup_uh;
  struct wait_queue *wait_queue; /* users waiting on select call */
  struct tq_struct psn;          /* queueing notifier to unmap   */
                                 /* whilst descheduling          */
  int int_vec;               /* the actual interrupt vector      */
  int irq;                   /* the irq line used by this PCIADA */
  int in_rw;                 /* flag for arranging acks in case  */
                             /* VMEbus errors                    */
  struct semaphore sem;      /* semaphore for SMP safety         */
} ArwvmeHardwareHandle;


#ifdef SEM_ENABLED
#define LOCK_HH(hh) down_interruptible(&hh->sem);
#define RELEASE_HH(hh) up(&hh->sem);
#else
#define LOCK_HH(hh) {} 
#define RELEASE_HH(hh) {} 
#endif

/*
 * there is a bug in the PLX9050:
 * the bit 7 of PCIBAR0/1 are
 * eventually set -> a PCI master
 * cannot read the LCR!!
 */
#define ARWVME_PCIBAR0(a)        ((a) & 0xFFFFFF80)
#define ARWVME_PCIBAR1(a)        ((a) & 0xFFFFFF80)
#define ARWVME_PCIBAR2(a)        ((a) & 0xFFFFFFF0) 

#define ARWVME_LAS0RR(a)         (((~(a)) & 0x0FFFFFF0)+0x10)

#define ARWVME_INTCSR(base)    (base + 0x4c)
#define ARWVME_CNTRL(base)     (base + 0x50)
#define ARWVME_EN_VMEMM(base)  writew_reg(0x4184, ARWVME_CNTRL(base))
#define ARWVME_DIS_VMEMM(base) writew_reg(0x4084, ARWVME_CNTRL(base))
#define ARWVME_VMEMM_OK(base)  (readw_reg(ARWVME_CNTRL(base)) & 0x800)

#define ARWVME_CLR_INT2(base)  ARWVME_DIS_VMEMM(base); \
                               ARWVME_EN_VMEMM(base)

/* globally enable/disable Interrupts (local1/2+pci) */
#define ARWVME_EN_INT(base)   writew_reg(0x49, ARWVME_INTCSR(base)) 
#define ARWVME_DIS_INT(base)  writew_reg(0x0, ARWVME_INTCSR(base)) 

/* Interrupt sources */
#define ARWVME_TST_PCIADA_INT(base) \
                                  (readw_reg(ARWVME_INTCSR(base))&0x20) 
#define ARWVME_TST_VMEMM_INT(base) \
                                  (readw_reg(ARWVME_INTCSR(base))&0x4) 
#define ARWVME_TST_INT(base) \
                                  (readw_reg(ARWVME_INTCSR(base))&0x24)
                                  

#define ARWVME_VME_BASE(mm_base)  (mm_base + 0x1000)

#define ARWVME_VEC_BASE(mm_base)  (mm_base + 0x800)

#define ARWVME_VIC_BASE(mm_base)  (mm_base + 0x400)

/*
 * reads and writes VMEADDRReg
 * (only A12..A31)
 */
#define ARWVME_WR_VMEADDR_REG_HW(mm_base,a)  (writew(a, mm_base + 0xA))
#define ARWVME_WR_VMEADDR_REG_LW(mm_base,a)  (writew(a, mm_base + 0x8))
#define ARWVME_WR_VMEADDR_REG_L(mm_base,a)   (writel(a, mm_base + 0x8))

#define ARWVME_RD_VMEADDR_REG_HW(mm_base)   (readw(mm_base + 0xA) << 16)
#define ARWVME_RD_VMEADDR_REG_LW(mm_base)   (readw(mm_base + 0x8))
#define ARWVME_RD_VMEADDR_REG_L(mm_base)    (readl(mm_base + 0x8))

#define ARWVME_VIC_GLOBAL_RESET(mm_base)    (writew(0xA, mm_base + 0x4))
#define ARWVME_VIC_INTERN_RESET(mm_base)    (writew(0x5, mm_base + 0x4))

#define ARWVME_INT_PENDING(mm_base)         (readw(mm_base + 0x4)) 
#define ARWVME_READ_INT_VEC(mm_base)        (readb(mm_base + 0x0800 + readw(mm_base + 0x4)))

#define ARWVME_INIT_RMC(mm_base)            (writew(0x1, mm_base))
#define ARWVME_TEST_RMC(mm_base)            (readw(mm_base)&0x1) 

/* Word-Path Jumper */
#define ARWVME_TEST_J401(mm_base)           (readw(mm_base)&0x4)
/* Systemcontroller Jumper */
#define ARWVME_TEST_J402(mm_base)           (readw(mm_base)&0x8)
/* Module number */
#define ARWVME_GET_MODULE_NUMBER(mm_base)   ((readw(mm_base)&0xF0)>>4)
#define ARWVME_GET_FPGA_REVISION(mm_base)   ((readw(mm_base)&0xF00)>>8)
/*
 * ModuleType should be 0x1 for MM 
 */
#define ARWVME_GET_MODULE_TYPE_ID(mm_base)  ((readw(mm_base)&0xF000)>>12)

/*
 * reads and writes VMEVICReg (VIC86A)
 */
#define ARWVME_RD_ADDR_MODIFIER(vic_base)   (readb(vic_base + VIC68_AMSR)&VIC68_AMSR_MASK)
#define ARWVME_WR_ADDR_MODIFIER(vic_base,a) (writeb((a)&VIC68_AMSR_MASK, vic_base + VIC68_AMSR))
/*
 * Reset of SYSFAIL always necessary 
 * after global Reset
 */
#define ARWVME_TEST_ICR7_7(vic_base)  (readb(vic_base + VIC68_ICR(6))&(1<<6))
#define ARWVME_CLR_SYSFAIL(vic_base)  B_SETBIT(vic_base + VIC68_ICR(7), 7) 

/*
 * set VMEBus Timeout
 */
#define ARWVME_TEST_TIMEOUT(vic_base) (readb(vic_base + VIC68_TTR)&0x2)

/*
 * Local Interrupt Control Register 
 * and Local Interrupt Vector Base Register
 * Reset-Key on Front = LICR6
 * VMEBus Timeout = LICR7
 */

#define ARWVME_LICR_SETUP(vic_base,lev,i)  (writeb(0x30|(lev&0x7), vic_base + VIC68_LICR(i))) 
#define ARWVME_LICR_ENABLE(vic_base,i)  B_SETBIT(vic_base + VIC68_LICR(i), 7) 
#define ARWVME_LICR_DISABLE(vic_base,i) B_CLRBIT(vic_base + VIC68_LICR(i), 7) 
#define ARWVME_LIRQ_STATUS(vic_base,i)  (!readb((vic_base + VIC68_LICR(i)&0x4)))

#define ARWVME_CLR_BESR(vic_base)    (writeb(0x0, vic_base + VIC68_BESR))
#define ARWVME_GET_BUS_ERR(vic_base) (readb(vic_base + VIC68_BESR)&VIC68_BERR_MASK)
#define ARWVME_GET_BUS_TO(vic_base)  (readb(vic_base + VIC68_BESR)&VIC68_VMETO_MASK)

#define ARWVME_BERR 1
#define ARWVME_BTO  2

#endif
