
/*
 * File   : test_mix.c
 *
 * Purpose: Test program for ARWVME interface kit
 *          testing the V462 GATE 
 *
 *
 * $Revision: 1.4 $
 * $Date: 2000/02/25 16:55:00 $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sched.h>
#include <arwvme.h>
#include <time.h>

#include <sys/times.h>

#define TTT 10
#define GATE_BASE 0x00FA0000
#define GATE_RANGE 0xFF

#define SWAP_WORD(a) (((a >> 16)) | ((a & 0xFFFF) << 16))

#ifndef TEST_MMAP
# define TEST_MMAP
#endif
#define TEST_INTERRUPT

static int arwvme_fd = -1;

unsigned int getBCD(int n)
{
  int i;
  unsigned int bcd = 0;
  for (i=0;i<2*sizeof(int);i++)
  {
    bcd += (n%10) << (i<<2);
    n /= 10;
    if (!n) break;
  }
  return bcd;
}

void sig_handler(int sig)
{
  ArwvmeIntAck *int_ack;
  ArwvmeSysErrInfo *sei;
  switch (sig) {
    case SIGUSR1:
      puts("sig_handler::SIGUSR1\n");
      int_ack = ArwvmeGetNextPendingInt(arwvme_fd);
      if (int_ack) {
        printf("Int 0x%x occured at %Ld\n",int_ack->int_vec,int_ack->int_tl);
        free(int_ack);
      }
    break;
    case SIGBUS:
      sei = ArwvmeGetSysErrInfo(arwvme_fd);
      if (errno) perror ("ArwvmeGetSysErrInfo");
      if (sei->module_number > 0) {
       printf("VMEBus Sys error: reason = %d, tl = %ld, addr = 0x%lx\n", sei->reason, sei->jiffies, sei->vme_addr);
      } else {
        printf("Buserror sent from kernel!\n");
        exit(1);
      }
    break;
    default:
      puts("sig_handler::default");
  }
}

/* CSR access via mmap */
#define derefCSR(base)     *(unsigned short *)base
#define GetError           (derefCSR(base) & 0x8000)
#define TestChanInVmeMode0 (derefCSR(base) & 0x2000)
#define TestChanInVmeMode1 (derefCSR(base) & 0x4000)
#define TestGatePresent0   (derefCSR(base) & 0x0800)
#define TestGatePresent1   (derefCSR(base) & 0x1000)
#define StartGate0         (derefCSR(base) |= 0x0200)
#define StartGate1         (derefCSR(base) |= 0x0400)

#define getVersion(base)   ((*(unsigned short *)(base+0xFE) & 0xF000) >> 12)
#define getSeries(base)    (*(unsigned short *)(base+0xFE) & 0x0FFF)
#define getManufact(base)  ((*(unsigned short *)(base+0xFC) & 0xFE00) >> 10)
#define getModType(base)   (*(unsigned short *)(base+0xFC) & 0x003F)       
#define identifyV462(base) (*(unsigned short *)(base+0xFC) == 0x80A)

/*
 * only A16 mode
 */
#define VMDIS_BASE      0xF000
#define VMDIS_START     0x0F00
#define VMDIS_RANGE     0xFF

/* Display access via mmap */
#define VMEDATL  0x0
#define VMEDATH  0x2
#define VMEADDRL 0x4
#define VMEADDRH 0x6
#define VMESTA0  0x8
#define VMESTA1  0xA
#define CSR0     0x1C
#define CSR1     0x1E

#define GET_VME_DATA(base)  *(unsigned short *)((char *)base+VMEDATL) | \
                            (*(unsigned short *)((char *)base+VMEDATH) << 16)
#define GET_VME_ADDR(base)  *(unsigned short *)((char *)base+VMEADDRL) | \
                            (*(unsigned short *)((char *)base+VMEADDRH) << 16)
#define GET_VMESTA0(base)   *(unsigned short *)((char *)base+VMESTA0)
#define GET_VMESTA1(base)   *(unsigned short *)((char *)base+VMESTA1)
#define GET_CSR0(base)      *(unsigned short *)((char *)base+CSR0)
#define GET_CSR1(base)      *(unsigned short *)((char *)base+CSR1)

void showModeStatus(char *base)
{
  unsigned short csr0 = GET_CSR0(base);
  switch (csr0 & 0xE) {
    case 2:
      printf("LATched Mode\n");
    break;
    case 4:
      printf("MONitor Mode\n");
    break;
    case 8:
      printf("DIRect Mode\n");
    break;
    default: return;
  }
}

void showLastData(char *base)
{
  printf("Last data: 0x%x\n",GET_VME_DATA(base));
}

void showLastAddr(char *base)
{
  printf("Last addr: 0x%x\n",GET_VME_ADDR(base));
}

void showTriggerLevelAndID(char *base)
{
  unsigned short csr1 = GET_CSR1(base);
  unsigned short vme_vector = csr1 & 0xFF;
  unsigned short vme_level = (csr1 & 0x700)>>8;
  unsigned int vme_int_pending = csr1 & 0x8000;
  printf("VME VECTOR: 0x%X\n",vme_vector);
  printf("VME LEVEL:  %d\n",vme_level);
  printf("INT PENDING: ");
  vme_int_pending ? printf("YES\n") : printf("NO\n"); 
}

void triggerInterrupt(char *base)
{
  unsigned short *_csr0 = (unsigned short *)(base + CSR0);
  unsigned short csr0 = *_csr0 & ~0xF;
  *_csr0 = csr0 | (1<<6);
  *_csr0 = csr0;  
}

int readV462InfosViaMMap(unsigned char *base, char *buf)
{
  int p = 0;
  p = sprintf(buf, "V462 Module Identifier Words\n");
  p += sprintf(buf+p, "Version: 0x%x, Serial Number: 0x%x\n",
                    getVersion(base),getSeries(base));
  p += sprintf(buf+p, "Manufacturer Nr.: 0x%x, Type: 0x%x\n",
                    getManufact(base), getModType(base));
  p += sprintf(buf+p, identifyV462(base) ? "Module is V462\n" : "Module unknown\n");
  return p;
}

void startGate0(unsigned int offs)
{
  unsigned short csr;
  lseek(arwvme_fd, offs+2, 0);
  read(arwvme_fd, &csr, 2);
  csr |= 0x200;
  lseek(arwvme_fd, offs+2, 0);
  write(arwvme_fd, &csr, 2);
}

void stopGate0(unsigned int offs)
{
  unsigned short csr;
  lseek(arwvme_fd, offs+2, 0);
  read(arwvme_fd, &csr, 2);
  csr &= ~0x200;
  lseek(arwvme_fd, offs+2, 0);
  write(arwvme_fd, &csr, 2);
}

void writeGateViaMMap(unsigned short *base, unsigned int usec, int chan)
{
  if (chan < 0 || chan > 1) return;
  *(base+2+2*chan) = usec & 0xFFFF;
  *(base+1+2*chan) = (usec & 0xFFFF0000) >> 16;
}

void writeGate(unsigned int offs, unsigned int usec, int chan)
{
  if (chan < 0 || chan > 1) return;
  lseek(arwvme_fd, offs+2+chan * 4, 0);
  write(arwvme_fd, &usec, 4);
}

int readGateViaMMap(unsigned short *base, int chan)
{
  unsigned int bcd;
  bcd = *(base + 2+ chan*2);
  bcd |= *(base + 1 + chan*2) << 16;
  return bcd;
}

int readGate(unsigned int offs, int chan)
{
  unsigned int gate;
  if (chan < 0 || chan > 1) return 0xFFFFFFFF;
  lseek(arwvme_fd, offs+2+chan*4, 0);
  read(arwvme_fd, &gate, 4);
  return gate; 
}

int readV462Infos(unsigned int offs, char *buf)
{
  unsigned short at_0xfc = 0;
  unsigned short at_0xfe = 0;
  lseek(arwvme_fd, offs+0xFC, 0);
  read(arwvme_fd, &at_0xfc, 2);
  lseek(arwvme_fd, offs+0xFE, 0);
  read(arwvme_fd, &at_0xfe, 2);
  printf("at_0xfc: 0x%x\n",at_0xfc);
  printf("at_0xfe: 0x%x\n",at_0xfe);
}

void main ( int argc, char *argv[] )
 {
  unsigned char *ret,buf[1024];
  int p,i;
  unsigned short a;
  time_t start;
  int offset = 0;
  int channel = 0;
  ArwvmeSysErrInfo *bei;
  unsigned short arr[9999];
  
  if (argc < 2) channel = 0;
  else channel = atoi(argv[1]);
  if (channel < 0 || channel > 1) channel = 0;
  printf ( "ArwvmeOpen: %d %d\n", arwvme_fd = ArwvmeOpen(2), errno );
  if (arwvme_fd < 0) exit(1);

  printf ( "ArwvmeCheckConnection: %d\n", ArwvmeCheckConnection(arwvme_fd) );
  
/* this is the only possible mode of the v462 */

#ifdef TEST_INTERRUPT
  ArwvmeEnableInterrupt(arwvme_fd);
  /* register interrupt handler and enable int */

    printf ( "Testing interrupts ...\n" );

    ArwvmeEnableInterrupt(arwvme_fd);
    /* very important: choose the correct AckType for your board! */ 
    ArwvmePostInterrupt(arwvme_fd, ArwvmeRoak, 0x0, 0x40, 0x0);
    /* AckType, Rora Register, Vector */

#if 0
    ArwvmeSignal(arwvme_fd, sig_handler);
#endif
/* catch the VME system errors */
    signal(SIGBUS, sig_handler); 
#endif

#ifdef TEST_MMAP
  /* --- test for mmap() call --- */
  ret = ArwvmeMMap(arwvme_fd, 0x1000);
  printf("ret: %p\n",ret);
  /* choose appropriate page on VME bus */
  if (ret) {
    /* accessing VMDIS mmapped */
    ArwvmeSetVmeWindow(arwvme_fd, VMDIS_BASE);
    i = 0;
  ArwvmeSetMode(arwvme_fd, 
                ARWVME_A24 | ARWVME_D16 | ARWVME_DATA | ARWVME_NONPRIV);
    for (;;) {
      writeGate(GATE_BASE, getBCD(i++), channel);
  ArwvmeSetMode(arwvme_fd, 
                ARWVME_A16 | ARWVME_D16 | ARWVME_DATA | ARWVME_NONPRIV);
      showLastData(ret+VMDIS_START);
      showModeStatus(ret+VMDIS_START);
      showTriggerLevelAndID(ret+VMDIS_START);
      showLastAddr(ret+VMDIS_START);
  ArwvmeSetMode(arwvme_fd, 
                ARWVME_A24 | ARWVME_D16 | ARWVME_DATA | ARWVME_NONPRIV);
      printf("readGate: %d\n", readGate(GATE_BASE, channel));
      sleep(1);
    }
  } else {
    printf("Sorry, no remap!\n");
  }
#endif
  ArwvmeShutdown(arwvme_fd);
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
