
/*
 * File   : test_vmdis.c
 *
 * Purpose: Test program for ARWVME interface kit
 *          testing the VME DISPLAY
 *
 *
 * $Revision: 1.5 $
 * $Date: 1999/10/07 09:16:02 $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sched.h>
#include <arwvme.h>
#include <time.h>

#include <sys/times.h>

/*
 * only A16 mode
 */
#define VMDIS_BASE      0xF000
#define VMDIS_START     0x0F00
#define VMDIS_RANGE     0xFF

#define SWAP_WORD(a) (((a >> 16)) | ((a & 0xFFFF) << 16))

static int arwvme_fd = -1;

void sig_handler(int sig)
{
  ArwvmeIntAck *int_ack;
  ArwvmeSysErrInfo *sei;
  switch (sig) {
    case SIGUSR1:
      puts("sig_handler::SIGUSR1");
      int_ack = ArwvmeGetNextPendingInt(arwvme_fd);
      if (int_ack) {
        printf("Int 0x%x occured at %Ld\n",int_ack->int_vec,int_ack->int_tl);
        free(int_ack);
      }
    break;
    case SIGBUS:
      puts("sighandler::SIGBUS");
      sei = ArwvmeGetSysErrInfo(arwvme_fd);
      if (sei->module_number > 0) {
       printf("VMEBus Sys error: reason = %d, tl = %ld, addr = 0x%lx\n", sei->reason, sei->jiffies, sei->vme_addr);
      } else {
        printf("Buserror sent from kernel!\n");
        exit(1);
      }
    break;
    default:
      puts("sig_handler::default");
  }
}

/* Display access via mmap */
#define VMEDATL  0x0
#define VMEDATH  0x2
#define VMEADDRL 0x4
#define VMEADDRH 0x6
#define VMESTA0  0x8
#define VMESTA1  0xA
#define CSR0     0x1C
#define CSR1     0x1E

#define GET_VME_DATA(base)  *(unsigned short *)((char *)base+VMEDATL) | \
                            (*(unsigned short *)((char *)base+VMEDATH) << 16)
#define GET_VME_ADDR(base)  *(unsigned short *)((char *)base+VMEADDRL) | \
                            (*(unsigned short *)((char *)base+VMEADDRH) << 16)
#define GET_VMESTA0(base)   *(unsigned short *)((char *)base+VMESTA0)
#define GET_VMESTA1(base)   *(unsigned short *)((char *)base+VMESTA1)
#define GET_CSR0(base)      *(unsigned short *)((char *)base+CSR0)
#define GET_CSR1(base)      *(unsigned short *)((char *)base+CSR1)

void showModeStatus(char *base)
{
  unsigned short csr0 = GET_CSR0(base);
  switch (csr0 & 0xE) {
    case 2:
      printf("LATched Mode\n");
    break;
    case 4:
      printf("MONitor Mode\n");
    break;
    case 8:
      printf("DIRect Mode\n");
    break;
    default: return;
  }
}

void showLastData(char *base)
{
  printf("Last data: %d\n",GET_VME_DATA(base));
}

void showLastAddr(char *base)
{
  printf("Last addr: 0x%x\n",GET_VME_ADDR(base));
}

void showTriggerLevelAndID(char *base)
{
  unsigned short csr1 = GET_CSR1(base);
  unsigned short vme_vector = csr1 & 0xFF;
  unsigned short vme_level = (csr1 & 0x700)>>8;
  unsigned int vme_int_pending = csr1 & 0x8000;
  printf("VME VECTOR: 0x%X\n",vme_vector);
  printf("VME LEVEL:  %d\n",vme_level);
  printf("INT PENDING: ");
  vme_int_pending ? printf("YES\n") : printf("NO\n"); 
}

void triggerInterrupt(char *base)
{
  unsigned short *_csr0 = (unsigned short *)(base + CSR0);
  unsigned short csr0 = *_csr0 & ~0xF;
  *_csr0 = csr0 | (1<<6);
  *_csr0 = csr0;  
}

void main ( int argc, char *argv[] )
 {
  unsigned char *ret;
  ArwvmeSysErrInfo *bei;
  const int ROUNDS = 1<<20;
  char *buf = malloc(ROUNDS);
  clock_t start;
  int i;
  unsigned short csr0;
  
  if (!buf) exit(1);
  if (argc < 3) {
    puts("Usage: test_vmdis major_number int_vec");
    exit(1);
  }
  printf ( "ArwvmeOpen: %d %d\n", arwvme_fd = ArwvmeOpen(atoi(argv[1])), errno );
  if (arwvme_fd < 0) exit(1);

  printf ( "ArwvmeCheckConnection: %d\n", ArwvmeCheckConnection(arwvme_fd) );
  
  ArwvmeSetMode(arwvme_fd, 
                ARWVME_A16 | ARWVME_D16 | ARWVME_DATA | ARWVME_NONPRIV);

#ifdef TEST_INTERRUPT
  ArwvmeEnableInterrupt(arwvme_fd);
  ArwvmePostInterrupt(arwvme_fd, ArwvmeRoak, 0x0, atoi(argv[2]), 0x0);
  ArwvmeSignal(arwvme_fd, sig_handler);
  signal(SIGBUS, sig_handler);
#endif
#ifdef TEST_MMAP
  /* --- test for mmap() call --- */
  ret = ArwvmeMMap(arwvme_fd, 0x1000);
  printf("ret: %p\n",ret);
  /* choose appropriate page on VME bus */
  if (ret) {
    ArwvmeSetVmeWindow(arwvme_fd, VMDIS_BASE);
    /* accessing VMDIS */
    showModeStatus(ret+VMDIS_START);
    showTriggerLevelAndID(ret+VMDIS_START);
    while (1) {
      showLastAddr(ret+VMDIS_START);
      showLastData(ret+VMDIS_START);
      sleep(1);
    }
  } else {
    printf("Sorry, no remap!\n");
  }
#endif
  ArwvmeUntieFpos(arwvme_fd);
  lseek(arwvme_fd, VMDIS_BASE+VMDIS_START+CSR0, 0);
  csr0 = 0x2;
  write(arwvme_fd, &csr0, 2);
  printf("csr0: 0x%x\n",csr0);
  lseek(arwvme_fd, VMDIS_BASE+VMDIS_START+VMEDATL, 0);
  start = times(NULL);
  i = ArwvmeRead(arwvme_fd, buf, ROUNDS);
  printf("time: %d\n",times(NULL)-start); 
  printf("%d bytes read\n",i);
  ArwvmeShutdown(arwvme_fd);
}

/* -------------------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
